'''
.. Created on: 2/12/15
   Common functions and data

.. codeauthor:: Rich Plevin <rich@plevin.com>

.. Copyright (c) 2015-2016 Richard Plevin
   See the https://opensource.org/licenses/MIT for license details.
'''
from ..constants import DEFAULT_TIMESTEP
from ..subcommand import SubcommandABC

class ChartCommand(SubcommandABC):
    def __init__(self, subparsers):
        kwargs = {'help' : '''Generate charts from CSV files generated by GCAM batch queries'''}

        super(ChartCommand, self).__init__('chart', subparsers, kwargs, group='project')

    def addArgs(self, parser):

        # argparse type checker/converter to handle symbolic unit conversions
        def floatOrConversion(string):
            import argparse
            from ..units import getUnits

            u = getUnits()
            try:
                return u.convert(string)
            except:
                raise argparse.ArgumentTypeError("%r is not a float or a known unit conversion" % string)

        # --byRegion and --region are mutually exclusive
        group1 = parser.add_mutually_exclusive_group()

        # --divisorFile and --divisor are mut. excl.
        group2 = parser.add_mutually_exclusive_group()

        # --sumYears and --valueCol are mut. excl.
        group3 = parser.add_mutually_exclusive_group()

        parser.add_argument('csvFile', nargs='?',
                            help='''The file containing the data to plot.''')

        parser.add_argument('-b', '--box', action="store_true",
                            help='''Draw a box around the plot. Default is no box.''')

        group1.add_argument('-B', '--byRegion', action="store_true",
                            help='''Generate one plot per region. Region names are read from the CSV file,
                            so they reflect any regional aggregation produced by the query.''')

        parser.add_argument('-c', '--columns', default="output",
                            help='''Specify the column whose values identify the segments in the stacked
                            bar chart. (These appear in the legend.)''')

        parser.add_argument('-C', '--constraint',
                            help='''Apply a constraint to limit the rows of data to plot. The constraint
                            can be any constraint string that is valid for the DataFrame.query() method,
                            e.g., -C 'input == "biomass"'
                            ''')

        parser.add_argument('-d', '--outputDir', default=".",
                            help='''The directory into which to write image files. Default is "."''')

        parser.add_argument('-D', '--workingDir', default='.',
                            help='''The directory to change to before performing any operations''')

        parser.add_argument('-e', '--enumerate', action="store_true",
                            help='''Prefix image filenames with sequential number for easy reference.
                            Used only with --fromFile''')

        parser.add_argument('-f', '--fromFile',
                            help='''A file from which to read argument strings, one per line.
                            These are read as if chartGCAM.py were called on each line individually,
                            but avoiding the ~2 sec startup time for the bigger python packages.''')

        group2.add_argument('-F', '--divisorFile',
                            help='''A file containing a floating point value to divide data by
                            before plotting. See also -V.''')

        parser.add_argument('--format',
                            help='''Specify a format for the Y-axis. Possible values are '.' for float,
                            ',' for int with commas, or any format recognized by print, e.g., "%%.2f" to
                            Y values as floats with 2 decimal places.''')

        parser.add_argument('-g', '--ygrid', action="store_true",
                            help="Show light grey horizontal lines at the major Y-axis ticks. Default is no grid.")

        parser.add_argument('-i', '--interpolate', action="store_true",
                            help="Interpolate (linearly) annual values between timesteps.")

        parser.add_argument('-I', '--indexCol', default="region",
                            help='''A column to use as the index column, or blank for None. This column
                            is displayed on the X-axis of stacked barcharts. Default value is "region".''')

        parser.add_argument('-k', '--yticks', action="store_true",
                            help="Show tick marks on Y-axis. Default is no tick marks.")

        parser.add_argument('-l', '--label', action="store_true",
                            help="Add text along the right side of the figure showing the filename.")

        parser.add_argument('-L', '--labelColor',
                            help='''Color for the text label, which defaults to lightgrey. Some users may
                            prefer "black", for example. (Implies -l)''')

        parser.add_argument('-m', '--multiplier', type=floatOrConversion,
                            help='''A value to multiply data by before generating the plot. The argument can
                            be a floating point number or the name of any variable in pygcam.unitConversion.py.
                            For example, "-m 3.667" and "-m C_to_CO2" are equivalent, and effectively convert
                            values from Tg C to Tg CO2. Be sure to adjust the Y axis label. See also -M.''')

        parser.add_argument('-M', '--multiplierFile',
                            help='''A file containing a floating point value to multiply data
                            by before plotting. See also -m.''')

        parser.add_argument('-n', '--ncol', type=int, default=5,
                            help='''The number of columns with which to display the legend. Default is 5.''')

        parser.add_argument('-N', '--scenario', default="",
                            help='''When using the '--fromFile' option, this argument is used to specify one
                            or more scenario names (delimited by commas if more than one). These are substituted
                            into each line read from the file as the value for "{scenario}" wherever it appears
                            on each line read from the 'fromFile'.''')

        parser.add_argument('--negate', action="store_true",
                            help="""Multiply data by -1 before plotting, which can make interpretation
                            of some figures more intuitive. The string "-negated" is added to the file
                            label, displayed if the "-l" or "-L" flag is specified.""")

        parser.add_argument('-o',  '--outFile', default="",
                            help='''The name of the image file to create. Format is determined from
                            filename extension. All common formats are allowed, e.g., png, pdf, tif,
                            and gif. Try it; it probably works. Default is the name of the data file
                            substituting ".png" for ".csv"''')

        parser.add_argument('-O', '--open', action="store_true",
                            help="Open the plot file after generating it.")

        parser.add_argument('-p',  '--palette', # "hls",
                            help='''The name of a color palette to use. Some good options include hls, husl, and Paired.
                            See http://stanford.edu/~mwaskom/software/seaborn/tutorial/color_palettes.html''')

        parser.add_argument('-r', '--rotation', type=int, default=90,
                            help='''Set the rotation angle for X-axis labels. Defaults to 90 degrees (vertical).
                            Use 0 for horizontal labels.''')

        parser.add_argument('-R', '--reference', default="reference",
                            help='''When using the '--fromFile' option, this argument is used to specify the
                            name of the reference scenario. The "other" scenario is given using the "-N" option.
                            These are substituted into each line read from the file as the value for "{scenario}"
                            and "{reference}" (without the quotes) wherever they appear on each line read
                            from the 'fromFile'. Defaults to "reference"''')

        group1.add_argument('--region',
                            help='''Plot values only for the given region.''')

        parser.add_argument('-s', '--skiprows', type=int, default=1,
                            help='''The number of rows of the CSV file to skip before reading the data (starting
                            with a header row with column names.) Default is 1, which works for GCAM batch
                            query output.''')

        group3.add_argument('-S', '--sumYears', action="store_true",
                            help='''Sum across the time horizon, typically by region. This results
                            in a stacked bar plot. When not summed over years (the default) a stacked area
                            plot is generated showing values grouped and summed by indexCol (-I) and
                            presented by year.''')

        parser.add_argument('-t', '--yearStep', type=int, default=DEFAULT_TIMESTEP,
                            help='''The spacing of year labels on X-axis for time-series plots.
                            Defaults to %d.''' % DEFAULT_TIMESTEP)

        parser.add_argument('-T', '--title', default="",
                            help='''Adds a title to the plot. Default is no title. The string can have
                            LaTeX math language in it, e.g., 'CO$_2$' causes the 2 to be subscripted, and
                            'MJ$^{-1}$' results in "MJ" with a superscripted exponent of -1. The string
                            '$\Delta$' results in a capital Greek delta. See LaTeX documentation for more
                            options. Be sure to enclose the title in single quotes so the "$" is not
                            (mis)interpreted by the shell.''')

        parser.add_argument('--timeseries', action='store_true',
                            help='''Plot the data as a time series.''')

        parser.add_argument('--transparent', action='store_true',
                            help='''Save the plot with a transparent background. (Default is white.)''')

        parser.add_argument('-u', '--unstackedCol',
                            help='''Draw an unstacked bar plot for the column given as an argument to this
                            option, showing three groups of bars: the region, all other regions, and the total.''')

        parser.add_argument('-U', '--unstackedRegion',
                            help='''The region to plot separately from Rest of World in an unstacked plot.
                            Ignored if --byRegion is specified, in which case a plot is created for all regions.''')

        group3.add_argument('-v', '--valueCol',
                            help='''Identify a single column (e.g., a year) to plot for bar plots.''')

        group2.add_argument('-V', '--divisor', type=floatOrConversion,
                            help='''A value to divide year column values by before plotting. The argument can
                            be a floating point number or the name of any variable in pygcam.unitConversion.py.
                            See also -F.''')

        parser.add_argument('-x', '--suffix',
                            help='''A suffix to append to the basename of the input csv file to create the
                            name for the output file. For example, if processing my_data.csv, indicating
                            -x '-by-region.pdf' results in an output file named my_data-by-region.pdf.''')

        parser.add_argument('-X', '--xlabel', default="",
                            help='''Defines a label for the X-axis; defaults to blank. LaTeX math language
                            is supported. (See the -T flag for more info.)''')

        parser.add_argument('-Y', '--ylabel', default="EJ",
                            help='''Label for the Y-axis; defaults to "EJ". LaTeX math language
                            is supported. (See the -T flag for more info.)''')

        parser.add_argument('-y', '--years', default="",
                            help='''Takes a parameter of the form XXXX-YYYY, indicating start and end
                            years of interest. Data for all other years are dropped.''')

        parser.add_argument('--ymax', type=float,
                            help='''Set the scale of a figure by indicating the value to show as the
                            maximum Y value. (By default, scale is set according to the data.)''')

        parser.add_argument('--ymin', type=float,
                            help='''Set the scale of a figure by indicating the value (given as abs(value),
                            but used as -value) to show as the minimum Y value''')

        parser.add_argument('-z', '--zeroLine', action="store_true",
                                help='''Whether to show a line at Y=0''')
        #
        # For manually tweaking figure layout
        #
        parser.add_argument('--legendY', type=float,
                            help='''The Y position of the legend. Useful for fixing poorly formatted figures.
                            Note that to pass a negative value, use the syntax --legendY="-xxx.xxx", otherwise
                            the hyphen is interpreted as indicating a command-line argument.''')

        parser.add_argument('--barWidth', type=float, default=0.5,
                            help='''The relative width of bars. Helpful when plotting only 1 or 2 bar, so they
                            aren't obnoxiously wide. Default is 0.5''')

        return parser


    def run(self, args, tool):
        from ..chart import chartMain    # avoid importing matplotlib and seaborn until needed

        chartMain(args, tool, self.parser)



# Copyright (c) 2016  Richard Plevin
# See the https://opensource.org/licenses/MIT for license details.

from ..error import PygcamMcsUserError
from .McsSubcommandABC import McsSubcommandABC

class ParallelPlotCommand(McsSubcommandABC):
    def __init__(self, subparsers):
        kwargs = {'help' : '''Generate a parallel coordinates plot for a set
        of simulation results. '''}
        super(ParallelPlotCommand, self).__init__('parallelPlot', subparsers, kwargs)
        self.group = None # don't display this in GUI for now

    def addArgs(self, parser):
        # Required arguments
        parser.add_argument('-r', '--resultName', type=str, required=True,
                            help='''The name of the result to create the plot for''')

        parser.add_argument('-s', '--simId',    type=int, required=True, help='The id of the simulation')

        parser.add_argument('-S', '--scenario', type=str, required=True, help='The name of the scenario')

        # Optional Arguments
        parser.add_argument('-b', '--inputBins', type=int, default=None,
                            help='''Allocate values for each variable into the given number of bins. By
                            default, the bins boundaries are evenly spaced. If the -q/--quantile flag
                            is given, the bins will contain an equal number of values. Use -l / --labels
                            to assign category names to the bins.''')

        parser.add_argument('-l', '--outputLabels', type=str, default='Low,Medium,High',
                            help='''Category names for the output bins. Value must be a comma-delimited
                            list of strings.''')

        parser.add_argument('--limit', type=int, default=0,
                            help='''Limit analysis to this number of trials''')

        parser.add_argument('-i', '--numInputs', type=int,
                            help='''The number of most-highly rank-correlated inputs to include in the figure.
                            By default, an attempt is made to plot all inputs.''')

        parser.add_argument('-I', '--invert', action='store_true',
                            help='''Plot negatively correlated data as (1 - x) rather than (x).''')

        parser.add_argument('-o', '--output', type=str,
                            help='''The name of the graphic output file to create. File format is determined from
                            the filename extension. Default is
                            {plotDir}/s{scenarioId}/{scenario}-{resultName}-parallel.png''')

        parser.add_argument('-q', '--quantiles', action='store_true',
                            help='''Create bins with an (approx.) equal number of values rather the
                            default, which is to space the bin boundaries equally across the range of values.''')

        parser.add_argument('-R', '--rotate', type=int, default=None,
                            help='''Angle of rotation for X-axis labels''')

        return parser   # for auto-doc generation


    def run(self, args, tool):
        from ..analysis import Analysis, makePlotPath

        simId = args.simId
        scenarioList = args.scenario.split(',')
        resultList = args.resultName.split(',')
        outputLabels = args.outputLabels.split(',')
        numInputs = args.numInputs
        bins = args.inputBins
        invert = args.invert
        quantiles = args.quantiles
        rotation = args.rotate

        anaObj = Analysis(args.simId, scenarioList, resultList, limit=args.limit)

        inputDF = anaObj.getInputs()

        # Drop any inputs with names ending in '-linked' since these are an artifact
        linked = filter(lambda s: s.endswith('-linked'), inputDF.columns)
        if linked:
            inputDF = inputDF.drop(linked, axis=1)

        if not (scenarioList and resultList):
            raise PygcamMcsUserError("scenario and resultName must be specified")

        resultDict = anaObj.getResults(scenarioList=scenarioList, resultList=resultList)

        for scenario in scenarioList:
            resultDF = resultDict[scenario]

            for resultName in resultList:
                resultSeries = resultDF[resultName]

                basename = '%s-s%02d-%s-parallel' % (scenario, simId, resultName)
                if bins:
                    basename += '-%d-bins' % bins
                if quantiles:
                    basename += '-quantiles'
                if invert:
                    basename += '-inverted'

                filename = makePlotPath(basename, simId)

                anaObj.plotParallelCoordinates(inputDF, resultSeries, inputBins=bins,
                                               numInputs=numInputs, outputLabels=outputLabels,
                                               invert=invert, rotation=rotation,
                                               quantiles=quantiles, filename=filename,
                                               extra=filename)



# Copyright (c) 2016  Richard Plevin
# See the https://opensource.org/licenses/MIT for license details.
#
# TBD: may be obsolete

from pygcam.log import getLogger
from .McsSubcommandABC import McsSubcommandABC

_logger = getLogger(__name__)

def driver(args, tool):
    '''
    Set up the base coremcs database. Must call startDb() before calling this.

    '''
    from ..Database import getDatabase

    db = getDatabase()
    expId = db.createExp(args.expName, args.description)
    _logger.debug("Added experiment '%s' with id=%d" % (args.expName, expId))


class AddExpCommand(McsSubcommandABC):
    def __init__(self, subparsers):
        kwargs = {'help' : '''Adds the named experiment to the database, with an optional description.'''}
        super(AddExpCommand, self).__init__('addexp', subparsers, kwargs)

    def addArgs(self, parser):
        parser.add_argument('-e', '--expName', type=str, required=True,
                            help='Add the named experiment to the database.')

        parser.add_argument('-d', '--description', type=str, required=False,
                            default='No description',
                            help='Add the named experiment to the database.')

        return parser   # for auto-doc generation


    def run(self, args, tool):
        driver(args, tool)



"""
.. codeauthor:: Rich Plevin <rich@plevin.com>

.. Copyright (c) 2016  Richard Plevin
   See the https://opensource.org/licenses/MIT for license details.
"""
#
# Generate template batch files, then launch ipyparallel
# controller and engines using the values in the template.
#
from __future__ import division, print_function
import copy
from six import iteritems
import os
import stat
import sys
from time import sleep
from IPython.paths import locate_profile

import ipyparallel as ipp
from ipyparallel.apps.ipclusterapp import ALREADY_STARTED, ALREADY_STOPPED, NO_CLUSTER

from .context import Context
from .Database import RUN_NEW, RUN_RUNNING, RUN_SUCCEEDED, RUN_QUEUED, RUN_KILLED, ENG_TERMINATE, getDatabase
from .error import IpyparallelError, PygcamMcsSystemError

from ..log import getLogger

# Exit values for Master.processTrials()
CONTINUE = 1
EXIT = 2

_logger = getLogger(__name__)

#
# SLURM
#
_slurmEngineBatchTemplate = """#!/bin/sh
#SBATCH --account={account}
#SBATCH --partition={queue}
#SBATCH --job-name={cluster_id}-engine
#SBATCH --nodes=1
#SBATCH --tasks-per-node={tasks_per_node}
#SBATCH --time={timelimit}
#{engine_args}
export MCS_WALLTIME={timelimit}
srun %s --profile-dir="{profile_dir}" --cluster-id="{cluster_id}" --prolog="{prolog_script}" --epilog="{epilog_script}"
"""
# These attempts accomplished nothing since TBB doesn't use them
# export OMP_NUM_THREADS=5
# export MKL_NUM_THREADS=1
# export MKL_DOMAIN_NUM_THREADS="BLAS=5"

# No need to run the controller on a compute node, but just in case
_slurmControllerBatchTemplate = """#!/bin/sh
#SBATCH --account={account}
#SBATCH --partition={queue}
#SBATCH --job-name={cluster_id}-controller
#SBATCH --ntasks=1
#SBATCH --time={timelimit}
%s --profile-dir="{profile_dir}" --cluster-id="{cluster_id}"
"""

#
# PBS
#
_pbsEngineBatchTemplate = """#!/bin/sh
#PBS -V
#PBS -N {cluster_id}-engine
#PBS -q {queue}
#PBS -l walltime={timelimit}
{engine_args}
%s --profile-dir="{profile_dir}" --cluster-id="{cluster_id}"
"""

_pbsControllerBatchTemplate = """#!/bin/sh
#PBS -V
#PBS -N {cluster_id}-controller
#PBS -q {queue}
#PBS -l walltime={timelimit}
%s --profile-dir="{profile_dir}" --cluster-id="{cluster_id}"
"""

#
# LSF
#
_lsfEngineBatchTemplate = """#!/bin/sh
TBD
"""

_lsfControllerBatchTemplate = """#!/bin/sh
TBD
"""

# TBD: test PBS (where?)
batchTemplates = {'slurm' : {'engine'     : _slurmEngineBatchTemplate,
                             'controller' : _slurmControllerBatchTemplate},

                  'pbs'   : {'engine'     : _pbsEngineBatchTemplate,
                             'controller' : _pbsControllerBatchTemplate},

                  'lsf'   : {'engine'     : _lsfEngineBatchTemplate,
                             'controller' : _lsfControllerBatchTemplate},
                  }

class Master(object):
    def __init__(self, args):
        self.args = args
        self.db = getDatabase(checkInit=False)
        self.client = None
        self.finished = False

        projectName = args.projectName

        # If we're just adding trials, no point in caching runs from database
        if not args.addTrials:
            # cache run definitions from the database and amend as necessary when creating runs
            for scenario in args.scenarios:
                rows = self.db.getRunInfo(args.simId, scenario, includeSucceededRuns=False)
                _logger.info('Caching info for %d runs of scenario %s', len(rows), scenario)
                for row in rows:
                    assert len(row) == 4, 'db.getRunInfo failed to return 4 values'
                    runId, simId, trialNum, status = row
                    context = Context(projectName=projectName, runId=runId, simId=simId,
                                      trialNum=trialNum, scenario=scenario, status=status)
                    #_logger.debug("Loaded %s", context)

    def waitForWorkers(self):
        from pygcam.config import getParamAsInt

        maxTries  = getParamAsInt('IPP.StartupWaitTries')
        seconds   = getParamAsInt('IPP.StartupWaitSecs')
        profile   = self.args.profile
        clusterId = self.args.clusterId
        client = None

        for i in range(1, maxTries+1):
            if client and len(client) > 0:
                return

            if client is None:
                try:
                    # default timeout is 10 seconds
                    self.client = client = ipp.Client(profile=profile, cluster_id=clusterId)

                except IOError:
                    if i == maxTries - 1:
                        raise

                    _logger.info("Waiting for client (%d/%d)", i, maxTries)
                    sleep(seconds)
                    continue

            if len(client) == 0:
                _logger.info("Waiting for engines (%d/%d)", i, maxTries)
                sleep(seconds)

        if not client or len(client) == 0:
            raise IpyparallelError('Failed to connect to engines')

    def queueTotals(self):
        """
        Return totals for queue status across all engines
        """
        dv = self.client[:]
        qstatus = dv.queue_status()

        totals = dict(queue=0, completed=0, tasks=0)

        for id, stats in iteritems(qstatus):
            if id == u'unassigned':
                totals[id] = stats
            else:
                for key, count in iteritems(stats):
                    totals[key] += count

        return totals

    def shutdownIdleEngines(self):
        '''
        Shutdown the engines.
        '''
        if len(self.client) == 0:
            _logger.info("No engines are running")
            return

        dview = self.client[:]
        status = dview.queue_status()

        # for key, value in iteritems(status):
        #     _logger.debug('  %s: %s', key, value)

        idleEngines = []

        unassigned = u'unassigned'

        # Shutdown idle engines if there are no unassigned tasks
        if status[unassigned] == 0:
            # collect ids of idle engines
            for id, stats in iteritems(status):
                if id == unassigned:    # ignore
                    continue

                # _logger.debug('%d: %s', id, stats)
                if stats[u'tasks'] == 0 and stats[u'queue'] == 0:
                    idleEngines.append(id)

        if idleEngines:
            _logger.info('Shutting down %d idle engines', len(idleEngines))
            self.client.shutdown(targets=idleEngines, block=True)

            maxTries = 4
            seconds  = 1

            expectedEngines = len(self.client) - len(idleEngines)
            for i in range(maxTries):
                if len(self.client) != expectedEngines:
                    sleep(seconds)

            # TBD: handle timeout waiting for engines to stop
            _logger.info("%d engines active", len(self.client))

    def createRuns(self, simId, scenario, trialNums):
        '''
        Create entries in the "run" table for the given simId, trialNums, and scenario.

        :param simId: (int) simulation ID
        :param scenario: (str) scenario name
        :param trialNums: (list of int) trial numbers
        :return: list of Context instances
        '''
        from .error import PygcamMcsUserError

        db = self.db
        session = db.Session()

        try:
            exp = db.getExp(scenario)
            if exp is None:
                raise PygcamMcsUserError("Unknown scenario '%s'" % scenario)

            expId    = exp.expId
            baseline = exp.parent

            runs = []
            projectName = self.args.projectName
            groupName   = self.args.groupName

            # Add a run record for each trial in this chunk
            for trialNum in trialNums:
                # Add a record in the "run" table listing this trial as "new"
                # (row for this simid, trialnum and expid is deleted if it exists)
                run = db.createRun(simId, trialNum, expId=expId, status=RUN_NEW, session=session)
                runs.append(run)

            session.commit()

        except Exception:
            raise

        finally:
            db.endSession(session)

        contexts = map(lambda r: Context(projectName=projectName, runId=r.runId,
                                         simId=simId, trialNum=r.trialNum,
                                         scenario=scenario, groupName=groupName,
                                         baseline=baseline, status=r.status), runs)
        return contexts

    def setRunStatuses(self, pairs):
        """
        Process a list of status changes in a single transaction, e.g., when setting
        the status for a long list of runs to "queued".
        """
        from .Database import getDatabase

        db = getDatabase()
        with db.sessionScope() as session:
            for context, status in pairs:
                self.setRunStatus(context, status=status, session=session)

    def setRunStatus(self, context, status=None, session=None):
        """
        Cache the status of this run, and if it has changed, save the new
        status to the database. Some context objects are retrieved from the
        worker tasks, so we lookup the equivalent in our local cache to test
        for whether a change has occurred.
        """
        debug = False   # set to True for extra diagnostic msgs

        if debug:
            _logger.debug('setRunStatus: %s', context)

        status = status or context.status

        cached = Context.getRunInfo(context.runId)
        if cached:
            if debug:
                _logger.debug('setRunStatus: cache hit: %s', cached)

            if cached.status == status:
                if debug:
                    _logger.debug('setRunStatus: no change; returning')
                return
        else:
            if debug:
                _logger.debug('adding context for runId %d to cache', context.runId)
            cached = context.saveRunInfo()

        _logger.info('%s -> %s', cached, status)
        cached.setVars(status=status)
        self.db.setRunStatus(context.runId, status, session=session)

    def runningTasks(self):
        from pygcam.utils import flatten

        qstatus = self.client.queue_status(verbose=True)
        ids = [rec['tasks'] for key, rec in iteritems(qstatus) if isinstance(key, (int, long))]
        return flatten(ids)

    def completedTasks(self):
        recs = self.client.db_query({'completed': {'$ne': None}}, keys=['msg_id'])
        ids = [rec['msg_id'] for rec in recs] if recs else None
        return ids

    def checkRunning(self):
        """
        Check for newly running tasks
        """
        tasks = self.runningTasks()
        # _logger.debug("Found %d running tasks", len(running))

        for task in tasks:
            try:
                ar = self.client.get_result(task, owner=False, block=False)

                # if tasks are ready, we handle them later.
                if ar.ready():
                    _logger.debug('checkRunning: skipping task with result (%s)', task)
                    continue

                if ar.data:
                    context = ar.data.get('context')
                    if context:
                        self.setRunStatus(context)

            except (KeyError, ipp.RemoteError) as e:
                _logger.warning('checkRunning: purging result: %s', e)
                self.client.purge_results(jobs=task)

            except Exception as e:
                _logger.warning("checkRunning: %s", e)

    def resubmit(self, task, context):
        _logger.info('Resubmitting task %s', context)
        self.client.resubmit(task)
        self.setRunStatus(context, RUN_QUEUED)

    def getResults(self, tasks):
        if not tasks:
            return None

        client = self.client
        results = []

        for task in tasks:
            try:
                ar = client.get_result(task, owner=True, block=False)

                try:
                    if not ar.ready():
                        continue

                except KeyError:        # stale message id can trigger KeyError
                    client.purge_results(jobs=task)
                    continue

                chunk = ar.get()
                workerResult = chunk[0]
                context = workerResult.context
                status = context.status

                if status == ENG_TERMINATE:
                    if ar.engine_id is not None:
                        _logger.info("Terminating engine %s: insufficient time remaining", ar.engine_id)
                        client.shutdown(ar.engine_id)
                        sleep(2)
                        self.resubmit(task, context)

                elif status == RUN_KILLED:
                    self.resubmit(task, context)
                    continue

                else:
                    results.append(workerResult)

            except Exception as e:
                # Raised if an engine dies, e.g., walltime expired.
                # With retries=1, should be able to recover from this.
                _logger.warning('getResults: %s', e)

            client.purge_results(jobs=task)

        return results

    def saveResults(self, results):
        '''
        Called on the master to save results to the database that were prepared by the worker.
        '''
        from .Database import getDatabase

        db = getDatabase()
        session = db.Session()

        try:
            # Delete all old values in first transaction
            for result in results:
                context = result.context
                resultsList = result.resultsList

                if resultsList:
                    # Delete any stale results for this runId (i.e., if re-running a given runId)
                    names = [resultDict['paramName'] for resultDict in resultsList]
                    ids = db.getOutputIds(names)
                    db.deleteRunResults(context.runId, outputIds=ids, session=session)

            db.commitWithRetry(session)

            # Add all new values in a second transaction
            for result in results:
                context = result.context
                resultsList = result.resultsList or []
                runId   = context.runId

                self.setRunStatus(context, session=session)

                if context.status != RUN_SUCCEEDED:
                    continue

                for resultDict in resultsList:
                    paramName  = resultDict['paramName']
                    value      = resultDict['value']
                    regionName = resultDict['regionName']

                    # Save the values to the database
                    if resultDict['isScalar']:
                        db.setOutValue(runId, paramName, value, session=session)
                    else:
                        regionId = db.getRegionId(regionName)   # cached; not a DB query
                        units = resultDict['units']
                        db.saveTimeSeries(runId, regionId, paramName, value, units=units, session=session)

            db.commitWithRetry(session)

        except Exception as e:
            session.rollback()
            # TBD: distinguish database save errors from data access errors?
            raise PygcamMcsSystemError("saveResults failed: %s" % e)

        finally:
            db.endSession(session)

    def checkCompleted(self):

        while True:
            completed = self.completedTasks()
            if not completed:
                return

            _logger.debug('%d completed tasks', len(completed))

            results = self.getResults(completed)
            if not results:
                _logger.warning('Purging %d completed tasks with no results (engine died?)', len(completed))
                self.client.purge_results(jobs=completed)
                continue

            # _logger.debug('Saving %d results', len(results))
            self.saveResults(results)
            # for result in results:
            #     self.saveResults(result)

            seconds = 2
            sleep(seconds)    # brief sleep before checking for more completed tasks

    def outstandingTasks(self):
        '''
        Return True if any tasks remain to be completed.
        '''
        tot = self.queueTotals()
        _logger.info('%d engines: %s' % (len(self.client), tot))

        return tot['tasks'] or tot['queue'] or tot['unassigned'] or self.completedTasks()

    def mainloop(self):
        """
        Takes parameters from arguments passed from runsim plugin.
        If `args.loopOnly` is False, run the trials identified in self.args.
        If `args.addTrials` is True, exit after running the trials, otherwise,
        loop until all tasks are completed, if "shutdown when idle" was given
        as a command-line arg, or loop indefinitely, allowing another task to
        add more trials to run.

        :return: none
        """
        from ipyparallel import NoEnginesRegistered
        from pygcam.error import PygcamException
        from .slurm import Slurm

        slurm = Slurm()

        args = self.args
        shutdownWhenIdle = args.shutdownWhenIdle

        if args.redoListOnly and args.statuses:
            listTrialsToRedo(self.db, args.simId, args.scenarios, args.statuses)
            return

        if not args.runLocal:
            self.waitForWorkers()    # wait for engines to spin up

        if not args.loopOnly:
            self.runTrials()

        if args.addTrials or args.runLocal:
            return      # Exit after adding or running trials

        while True:
            try:
                self.checkRunning()         # Check for newly running tasks
                self.checkCompleted()       # Check for completed tasks

                outstanding = self.outstandingTasks()
                if outstanding:
                    if len(self.client) == 0:
                        pending = slurm.jobsInState('pending', jobName='mcs-engine')
                        if pending:
                            _logger.info('No engines registered; %d workers PENDING', len(pending))
                            sleep(30)
                            continue
                        else:
                            _logger.info("No engines running or pending. Shutting down hub.")
                            self.client.shutdown(hub=True, block=True)
                            return

                    # self.checkRunning()         # Check for newly running tasks
                    # self.checkCompleted()       # Check for completed tasks

                else:   # no outstanding tasks
                    _logger.info("No outstanding tasks. Shutting down hub.")
                    self.client.shutdown(hub=True, block=True)
                    return

                if shutdownWhenIdle:
                    self.shutdownIdleEngines()

                _logger.debug('sleep(%d)', args.waitSecs)
                sleep(args.waitSecs)

            except NoEnginesRegistered:
                pass    # handled in loop

            except Exception as e:
                raise PygcamException("mainloop: %s" % e)

    def runTrials(self):
        from .util import parseTrialString
        import worker

        args = vars(self.args)
        argDict = {}
        for key in ('runLocal', 'noGCAM', 'noBatchQueries', 'noPostProcessor'):
            argDict[key] = args.get(key, False)

        simId       = args['simId']
        statuses    = args['statuses']
        scenarios   = args['scenarios']
        projectName = args['projectName']
        groupName   = args['groupName']
        trialStr    = args['trials']
        runLocal    = args['runLocal']

        asyncResults = []
        view = None if runLocal else self.client.load_balanced_view(retries=2)

        for scenario in scenarios:

            if statuses:
                # Change this to return Run instances?
                # If any of the "redo" options find trials, use these instead of args.trials
                contexts = self.db.getRunsByStatus(simId, scenario, statuses,
                                                   projectName=projectName,
                                                   groupName=groupName)

                if not contexts:
                    _logger.warn("No trials found for simId=%s, scenario=%s with statuses=%s",
                                 simId, scenario, statuses)
                    continue
            else:
                trialCount = self.db.getTrialCount(simId)

                if trialStr:
                    # convert arg string like "4,7,9-12,42" to a list of ints
                    trialList = parseTrialString(trialStr)
                    userTrials = len(trialList)

                    # remove nonsense values and warn user about them
                    trialNums = filter(lambda trial: 0 <= trial < trialCount, trialList)
                    goodTrials = len(trialNums)
                    if goodTrials != userTrials:
                        _logger.warn('Ignoring %d trial numbers that are out of range [0,%d]',
                                     userTrials - goodTrials, trialCount)
                else:
                    # if trials aren't specified, queue all of them
                    trialNums = range(trialCount)

                contexts = self.createRuns(simId, scenario, trialNums)

            statusPairs = []

            for context in contexts:
                try:
                    if runLocal:
                        #self.setRunStatus(context, status=RUN_RUNNING)
                        statusPairs.append((context, RUN_RUNNING))
                        ctx = copy.copy(context)    # use a copy to simulate what happens with remote call...
                        result = worker.runTrial(ctx, argDict)
                        self.saveResults([result])
                    else:
                        # Easier to deal with a list of AsyncResults instances than a
                        # single instance that contains info about all "future" results.
                        result = view.map_async(worker.runTrial, [context], [argDict])
                        #self.setRunStatus(context, status=RUN_QUEUED)
                        statusPairs.append((context, RUN_QUEUED))

                    asyncResults.append(result)

                except Exception as e:
                    _logger.error("Exception running 'runTrial': %s", e)

            self.setRunStatuses(statusPairs)

        return asyncResults


def listTrialsToRedo(db, simId, scenarios, statuses):
    from .util import createTrialString

    # 'missing' is not a real status found in the database
    missing = 'missing' in statuses
    if missing:
        statuses.remove('missing')

    for scenario in scenarios:
        trialNums = []

        # If any of the "redo" options find trials, use these instead of args.trials
        tuples = db.getRunsByStatus(simId, scenario, statuses)
        if tuples:
            # Just print the corresponding trial string
            trialNums += map(lambda tuple: tuple[2], tuples)  # unpack(runId, simId, trialNum, ...) tuple

        if missing:
            trialNums += db.getMissingTrials(simId, scenario)

        if trialNums:
            trialStr = createTrialString(trialNums)
            print("%s:\n%s" % (scenario, trialStr))
        else:
            _logger.warn("No trials found for simId=%s, scenario=%s with statuses=%s",
                         simId, scenario, statuses)

def templatePath(scheduler, profile, clusterId, process):
    profileDir = locate_profile(profile)
    basename = '%s_%s_%s.template' % (scheduler, clusterId, process)
    filename = os.path.join(profileDir, basename)
    return filename

def _saveBatchFiles(numTrials, argDict):
    """
    Create and save engine and controller batch files that will
    be used by ipcontroller and ipengine at start-up. The files are
    created in the profile directory, indicated in `argDict` either
    as "profile_dir" or constructed from "profile" by ipython's
    `locate_profile` function. The ipyparallel config files must set
    the batch_template_file to the names of the generated files, which
    are "{scheduler}-{cluster_id}-{process}.template",
    where {process} is "engine" or "controller", depending as appropriate.
    So for SLURM (the default scheduler), and the default cluster id ("mcs"),
    the two generated files are "slurm-mcs-engine.template" and
    "slurm-mcs-controller.template".

    :param numTrials: (int) the number of GCAM trials to run
    :param argDict: (dict) values for "scheduler", "account", "queue",
       "profile", "cluster_id", "tasks_per_node", "num_engines", "timelimit",
       and "minimum_seconds" are substituted into the batch templates. For
       most of these, defaults are taken from ~/.pygcam.cfg and overridden
       by argDict. Both "ntasks" and "timelimit" are computed on-the-fly.
    :return: (dict) the names of the generated files, keyed by 'engine' or 'controller'
    """
    import pipes
    from pygcam.config import getParam, getParamAsInt

    minutesPerRun = argDict['minutesPerRun']
    maxEngines    = argDict['maxEngines']
    numEngines    = min(numTrials, maxEngines)
    maxRunsPerEngine = numTrials // numEngines + (1 if numTrials % numEngines else 0)

    minutesPerEngine = minutesPerRun * maxRunsPerEngine
    timelimit = "%02d:%02d:00" % (minutesPerEngine // 60, minutesPerEngine % 60)

    defaults = {'scheduler'       : getParam('IPP.Scheduler'),
                'account'         : getParam('IPP.Account'),
                'queue'           : getParam('IPP.Queue'),
                'engine_args'     : getParam('IPP.OtherEngineArgs'),
                'cluster_id'      : argDict['clusterId'],
                'tasks_per_node'  : getParamAsInt('IPP.TasksPerNode'),
                'min_secs_to_run' : getParamAsInt('IPP.MinTimeToRun') * 60,
                'timelimit'       : timelimit,
                'prolog_script'   : getParam('IPP.PrologScript'),
                'epilog_script'   : getParam('IPP.EpilogScript'),
                }

    defaults.update(argDict)

    profile    = defaults['profile']
    scheduler  = defaults['scheduler']
    cluster_id = defaults['cluster_id']

    defaults['profile_dir'] = locate_profile(profile)

    templates = batchTemplates[scheduler.lower()]
    files = {}

    # N.B. process is 'controller' or 'engine'
    for process, template in templates.items():
        cmd_argv = [sys.executable, "-m", "ipyparallel.%s" % process]
        text = template % ' '.join(map(pipes.quote, cmd_argv))  # insert command line
        text = text.format(**defaults)   # insert other parameters

        files[process] = filename = templatePath(scheduler, profile, cluster_id, process)

        _logger.debug('Writing %s', filename)
        with open(filename, 'w') as f:
            f.write(text)

        os.chmod(filename, stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR)  # make executable

    return files


def _clusterCommand(cmd):
    from pygcam.utils import shellCommand

    status = shellCommand(cmd, shell=True, raiseError=False)

    if status:
        statusStrings = {
            ALREADY_STOPPED: 'Cluster already stopped',
            ALREADY_STARTED: 'Cluster already started',
            NO_CLUSTER:      'No cluster found',
        }
        msg = statusStrings.get(status, 'Exit status: %d' % status)
        _logger.warning(msg)

    return status


def startEngines(numTrials, batchTemplate):
    """
    Uses the batch file created when the cluster was started, so it has
    the profile, cluster-id, and ntasks-per-node already set.
    """
    from pygcam.config import getParamAsInt
    from pygcam.utils import shellCommand

    tasksPerNode = getParamAsInt('IPP.TasksPerNode')
    maxEngines   = getParamAsInt('IPP.MaxEngines')
    numEngines   = min(numTrials, maxEngines)
    numNodes     = numEngines // tasksPerNode + int(bool(numEngines % tasksPerNode)) # rounds up

    # TBD: use "ipcluster engines" if there's a way to avoid leaving procs running after shutdown
    # enginesCmd = "ipcluster engines -n %d" % tasksPerNode
    enginesCmd = "sbatch '%s'" % batchTemplate

    for i in range(numNodes):
        _logger.info(enginesCmd)
        status = shellCommand(enginesCmd, shell=True, raiseError=False)
        if status:
            _logger.error('Attempt to launch engines failed; status=%d', status)
            return


def _defaultFromConfig(argDict, name, param):
    '''
    If kwargs doesn't have the named keyword, default to the value of
    the config file param, and set the kwargs dict so this value will
    be used in formatting the template.
    '''
    from pygcam.config import getParam

    if argDict.get(name, None) is None:
        argDict[name] = getParam(param)

    return argDict[name]

def pidFileExists(profile, clusterId):
    """
    Check if a process-id (pid) file exists for this profile and clusterId.

    :param profile: (str) profile name
    :param clusterId: (str) cluster ID
    :return: (bool) True if computed pid file pathname exists, else False
    """
    profileDir = locate_profile(profile)
    basename = 'ipcluster-%s.pid' % clusterId
    pidFile = os.path.join(profileDir, 'pid', basename)
    return os.path.isfile(pidFile)


def startCluster(**kwargs):
    """
    Start an ipyparallel cluster to run `numTrials`, using the given `profile`
    and `cluster_id`, if specified, otherwise using values of config vars
    IPP.Profile and IPP.ClusterId, respectively. Writes out a template batch
    file for use with either SLURM or PBS, depending on settings in the user's
    .pygcam.cfg, which is then used by the ipcluster command when launching
    the controller and engines. If the number of trials to run (`numTrials`)
    is greater than the value of config variable IPP.MaxEngines, this max
    number of engines is run but the time limit per engine is increased to allow
    engine each to run the required number of tasks.

    :param clusterId: (str) an id string for the new cluster
    :param numTrials: (int) the number of GCAM trials to run
    :param otherArgs: (str) other arguments appended to the ipcluster command-line
    :param profile: (str) the ipython profile to use
    :param workDir: (str) a directory to change to before starting the cluster
    :return: (int) exit status from the ipcluster command (zero => success)
    """
    # :param maxEngines: (int) the maximum number of engines that will run at once
    # :param minutesPerRun: (int) the time limit to set for each GCAM trial
    # :param queue: (str) the name of the queue to submit the jobs on
    # _defaultFromConfig(kwargs, 'queue',         'IPP.Queue')
    # _defaultFromConfig(kwargs, 'maxEngines',    'IPP.MaxEngines')

    clusterId = _defaultFromConfig(kwargs, 'clusterId', 'IPP.ClusterId')
    profile   = _defaultFromConfig(kwargs, 'profile',   'IPP.Profile')
    _defaultFromConfig(kwargs, 'minutesPerRun', 'IPP.MinutesPerRun')
    _defaultFromConfig(kwargs, 'otherArgs', 'IPP.OtherClusterArgs')
    _defaultFromConfig(kwargs, 'workDir', 'IPP.WorkDir')

    numTrials = kwargs['numTrials']
    templates = _saveBatchFiles(numTrials, kwargs)
    sleep(1)    # allow files to flush and close

    # If the pid file exists, stop the cluster first
    if pidFileExists(profile, clusterId):
        _logger.info('Stopping cluster "%s"', clusterId)
        stopCluster(profile=profile, cluster_id=clusterId, stop_jobs=True)
        sleep(1)

    # Start cluster with no engines; we add them later based on trial count
    template = "ipcluster start -n 0 --daemonize --profile={profile} --cluster-id={clusterId} --work-dir={workDir} {otherArgs}"
    controllerCmd = template.format(**kwargs)
    status = _clusterCommand(controllerCmd)

    if status == 0:
        startEngines(numTrials, templates['engine'])

    return status


def stopCluster(profile=None, cluster_id=None, stop_jobs=False, other_args=None):
    from pygcam.config import getParam
    from pygcam.utils import shellCommand

    # This allows user to pass empty string (e.g., -c='') to override default
    cluster_id = getParam('IPP.ClusterId') if cluster_id is None else cluster_id
    profile    = getParam('IPP.Profile')   if profile    is None else profile
    other_args = getParam('IPP.OtherClusterArgs') if other_args is None else other_args

    template = "ipcluster stop --profile={profile} --cluster-id={cluster_id} {other_args}"
    cmd = template.format(profile=profile, cluster_id=cluster_id, other_args=other_args)

    # shutdown engines, including those that cluster may not be tracking directly
    try:
        client = ipp.Client(profile=profile, cluster_id=cluster_id, timeout=1)
        client.shutdown(hub=False)
    except:
        pass

    status = _clusterCommand(cmd)

    # kill the engines
    if stop_jobs:
        cmd = getParam('IPP.StopJobsCommand').strip()
        if cmd:
            shellCommand(cmd, shell=True, raiseError=False)

    return status


# if __name__ == '__main__':



"""
.. Support for querying GCAM's XML database and processing results.

.. Copyright (c) 2016 Richard Plevin
   See the https://opensource.org/licenses/MIT for license details.
"""
import os
import re
import subprocess

from lxml import etree as ET

from .Xvfb import Xvfb
from .config import getParam, getParamAsBoolean, getParamAsFloat
from .constants import NUM_AEZS, GCAM_32_REGIONS
from .error import PygcamException, ConfigFileError, FileFormatError, CommandlineError, FileMissingError
from .log import getLogger
from .queryFile import QueryFile, RewriteSetParser
from .utils import (mkdirs, deleteFile, ensureExtension, ensureCSV, parse_version_info,
                    pathjoin, saveToFile, getExeDir, writeXmldbDriverProperties)
from .temp_file import TempFile, getTempFile

_logger = getLogger(__name__)

def limitYears(df, years):
    """
    Modify df to drop all years outside the range given by `years`.

    :param years: a sequence of two years (str or int); only values in this
        range (inclusive) are kept. Data for other years is dropped.
    :return: (DataFrame) df, modified in place.
    """
    first, last = map(int, years)
    yearCols  = map(int, filter(str.isdigit, df.columns))
    dropYears = map(str, filter(lambda y: y < first or y > last, yearCols))
    df.drop(dropYears, axis=1, inplace=True)
    return df

# TBD: simplify using this approach:

# def interp(df):
#     import numpy as np
#
#     years = sorted(map(int, filter(str.isdigit, df.columns)))
#     all_years = range(years[0], years[-1] + 1)
#     interp_yrs = sorted(set(all_years) - set(years))
#
#     # create the cols to interpolate and set to NaN for interpolate()
#     for y in interp_yrs:
#         df[str(y)] = np.nan
#
#     df = df.sort_index(axis=1).interpolate(axis=1)
#     return df


def interpolateYears(df, startYear=0, inplace=False):
    """
    Interpolate linearly between each pair of years in the GCAM output. The
    time-step is calculated from the numerical (string) column headings given
    in the `DataFrame`_ `df`, which are assumed to represent years in the time-series.
    The years to interpolate between are read from `df`, so there's no dependency
    on any particular time-step, or even on the time-step being constant.

    :param df: (DataFrame) Data of the format returned by batch queries
        on the GCAM XML database
    :param startYear: (int) If non-zero, begin interpolation at this year.
    :param inplace: (bool) If True, modify `df` in place; otherwise modify a copy.
    :return: if `inplace` is True, `df` is returned; otherwise a copy
      of `df` with interpolated values is returned.
    """
    yearCols = filter(str.isdigit, df.columns)
    years = map(int, yearCols)

    for i in range(0, len(years)-1):
        start = years[i]
        end   = years[i+1]
        timestep = end - start

        if timestep == 1:       # don't interpolate annual results if already annual
            continue

        startCol = df[str(start)]
        endCol   = df[str(end)]

        # compute vector of annual deltas for each row
        delta = (endCol - startCol)/timestep

        # interpolate the whole column -- but don't interpolate before the start year
        for j in range(1, timestep):
            nextYear = start + j
            df[str(nextYear)] = df[str(nextYear-1)] + (0 if nextYear < startYear else delta)

    yearCols = filter(str.isdigit, df.columns)  # get annualized year columns
    years = map(int, yearCols)       # sort as integers
    years.sort()
    yearCols = map(str, years)       # convert back to strings, now sorted

    nonYearCols = list(set(df.columns) - set(yearCols))
    result = df.reindex_axis(nonYearCols + yearCols, axis=1, copy=(not inplace))
    return result

_csvCache = {}

def readCsv(filename, skiprows=1, years=None, interpolate=False, startYear=0, cache=False):
    """
    Read a CSV file of the form generated by GCAM batch queries, i.e., skip one
    row and then read column headings and data. Optionally drop all years outside
    the `years` given. Optionally, linearly interpolate annual values between
    time-steps.

    :param filename: (str) the path to a CSV file
    :param skiprows: (int) the number of rows to skip before reading the data matrix
    :param years: (iterable of two values coercible to int) the year columns to
        keep; others are dropped
    :param interpolate: (bool) If True, interpolate annual values between time-steps
    :param startYear: (int) If interpolating, the year to begin interpolation
    :param cache: (bool) If True, file will be sought in, and saved to, a CSV cache.
       The "raw" file data is cached, so if called with different processing args,
       the same initial DataFrame is used, but it will be processed correctly.
    :return: (DataFrame) the data read in, processed as per arguments
    """
    import pandas as pd

    found = False

    if cache and filename in _csvCache:
        _logger.debug("Found %s in CSV cache", filename)
        df = _csvCache[filename]
        found = True

    else:
        try:
            _logger.debug("Reading %s", filename)
            df = pd.read_table(filename, sep=',', skiprows=skiprows, index_col=None)

        except IOError as e:
            raise FileMissingError(os.path.abspath(filename), e)

        except Exception as e:
            raise PygcamException('Error reading %s: %s' % (filename, e))

    # Cache a copy, unless it was already in the cache
    if cache and not found:
        _csvCache[filename] = df.copy()

    if years:
        limitYears(df, years)

    if interpolate:
        df = interpolateYears(df, startYear=startYear)

    return df

def writeCsv(df, filename, header='', float_format="%.4f"):
    """
    Save a DataFrame a file in "standard" GCAM csv format', which
    means without a numerical index, and with column headers on
    the second line.

    :param df: (pandas.DataFrame) a DataFrame holding the data to write
    :param filename: (str) the name of the file to create
    :param header: (str) a string to write as the first line of the file,
      (a single line preceding column headers is standard GCAM query result
      format.)
    :param float_format: (str) a format string indicating how to represent
      numeric values. Default shows 4 decimal places. To limit results to,
      for example, 2 decimal places, use float_format="%.2f".
    :return: none
    """
    _logger.info("Writing %s", filename)

    txt = df.to_csv(None, float_format=float_format)
    with open(filename, 'w') as f:
        f.write("%s\n" % header)  # add a header line to match batch-query output format
        f.write(txt)

# TBD: This belongs with gcamtool. Currently used only by constraints.py
def  readQueryResult(batchDir, baseline, queryName, years=None, interpolate=False, startYear=0, cache=False):
    """
    Compose the name of the 'standard' result file, read it into a DataFrame and
    return the DataFrame. Data is read from the computed filename
    "{batchDir}/{queryName}-{baseline}.csv".

    :param batchDir: (str) a directory in which the data file resides
    :param baseline: (str) the name of a baseline scenario
    :param queryName: (str) the name of a batch query.
    :param years: (iterable of two values coercible to int) the year columns to keep;
      others are dropped
    :param interpolate: (bool) If True, interpolate annual values between time-steps
    :param startYear: (int) If interpolating, the year to begin interpolation
    :param cache: (bool) If True, files will be sought in and saved to a CSV cache
    :return: (DataFrame) the data in the computed filename.
    """
    pathname = pathjoin(batchDir, '%s-%s.csv' % (queryName, baseline))
    df= readCsv(pathname, years=years, interpolate=interpolate, startYear=startYear, cache=cache)
    return df

# TBD: not used, but may be useful
def getRegionList(workspace=None):
    """
    Get a list of the defined region names.

    :param workspace: the path to a ``Main_User_Workspace`` directory that
      has the file
      ``input/gcam-data-system/_common/mappings/GCAM_region_names.csv``,
      or ``None``, in which case the value of config variable
      ``GCAM.SourceWorkspace`` (if defined) is used. If `workspace` is
      empty or ``None``, and the config variable ``GCAM.SourceWorkspace`` is
      empty (the default value), the built-in default 32-region list is returned.
    :return: a list of strings with the names of the defined regions
    """
    relpath = 'input/gcam-data-system/_common/mappings/GCAM_region_names.csv'

    workspace = workspace or getParam('GCAM.SourceWorkspace')
    if not workspace:
        return GCAM_32_REGIONS

    path = pathjoin(workspace, relpath)

    _logger.debug("Reading:", path)
    df = readCsv(path, skiprows=3)  # this is a gcam-data-system input file (different format)
    regions = list(df.region)
    return regions

def readRegionMap(filename):
    """
    Read a region map file containing one or more tab-delimited lines of the form
    ``key`` <tab> ``value``, where `key` should be a standard GCAM region and
    `value` the name of the region to map the original to, which can be an
    existing GCAM region or a new name defined by the user.

    :param filename: the name of a file containing region mappings
    :return: a dictionary holding the mappings read from `filename`
    """
    import re
    mapping = {}
    pattern = re.compile('\t+')

    _logger.info("Reading region map '%s'", filename)
    with open(filename) as f:
        lines = f.readlines()

    for line in lines:
        line = line.strip()
        if line[0] == '#':
            continue

        tokens = pattern.split(line)
        if len(tokens) != 2:
            raise FileFormatError("Badly formatted line in region map '%s': %s" % (filename, line))

        mapping[tokens[0]] = tokens[1]

    return mapping

def dropExtraCols(df, inplace=True):
    """
    Drop some columns that GCAM queries sometimes return, but which we generally don't need.
    The columns to drop are taken from from the configuration file variable ``GCAM.ColumnsToDrop``,
    which should be a comma-delimited string. The default value is ``scenario,Notes,Date``.

    :param df: a `DataFrame`_ hold the results of a GCAM query.
    :param inplace: if True, modify `df` in-place; otherwise return a modified copy.
    :return: the original `df` (if inplace=True) or the modified copy.
    """
    columns = df.columns
    # eliminate any extra (empty) columns that appear to be query artifacts
    dropCols = filter(lambda s: s.startswith('Unnamed:'), columns)

    colString = getParam('GCAM.ColumnsToDrop')
    colList = colString and colString.split(',')

    if colString and not colList:
        raise ConfigFileError("The value of %s is '%s'; should be a comma-delimited list of column names")

    unneeded = set(colList)
    existing = set(columns)
    dropCols += existing & unneeded    # drop any columns in both sets

    resultDF = df.drop(dropCols, axis=1, inplace=inplace)
    return resultDF

def _removeLevelByName(rewriteList, levelName):
    """
    If there was a hard-coded rewrite for levelName, delete it so we
    can use the one defined in the query file.
    """
    nodes = rewriteList.xpath("./level[@name='%s']" % levelName)
    for levelElt in nodes:
        rewriteList.remove(levelElt)

def _addRegionMap(regionMap, rewriteList):
    _removeLevelByName(rewriteList, 'region')

    levelElt = ET.Element('level', name='region')
    rewriteList.append(levelElt)

    for fromReg, toReg in regionMap.items():
        rewrite = ET.Element('rewrite', attrib={'from': fromReg, 'to': toReg})
        levelElt.append(rewrite)

def _addRewrites(levelElt, rewriteSet):

    def _appendRewrite(From, to):
        'Helper function to reduce redundancy'
        node = ET.Element('rewrite', attrib={'from': From, 'to': to})
        levelElt.append(node)

    byAEZ = rewriteSet.byAEZ
    byBasin = rewriteSet.byBasin

    for rewrite in rewriteSet.rewrites:
        From = rewrite.From             # "from" is a keyword, thus "From"
        to = rewrite.to

        if byAEZ:
            # Generate a rewrite for each AEZ
            for aez in range(1, NUM_AEZS + 1):
                fromAEZ = From + 'AEZ%02d' % aez
                _appendRewrite(fromAEZ, to)

        # TBD: not sure this makes sense here, actually...
        if byBasin:
            # Generate a rewrite for each basin-region combo? # TBD: GCAM5
            # for basin in ?
            #     fromBasin = 'From +
            #     _appendRewrite(fromAEZ, to)
            pass

        else:
            _appendRewrite(From, to)

def _addRewriteSet(rewriteSetList, rewriteParser, rewriteList, title):
    # TBD: Verify with Pralit.
    # Set to True if any of the mappings requires this, since it apparently we
    # cannot set these independently in different rewrite lists.
    appendValues = False

    for rewriteSetName, level in rewriteSetList:
        rewriteSet = rewriteParser.getRewriteSet(rewriteSetName)
        levelName = level or rewriteSet.level

        _removeLevelByName(rewriteList, levelName)
        levelElt = ET.Element('level', name=levelName)

        _logger.debug("Applying rewriteSet '%s', level '%s' for '%s'", rewriteSetName, levelName, title)

        rewriteList.append(levelElt)
        _addRewrites(levelElt, rewriteSet)
        appendValues = appendValues or rewriteSet.appendValues

    rewriteList.set('append-values', 'true' if appendValues else 'false')

def _findQueryByName(tree, title):
    """
    Try the title and variations thereof to locate the query by name
    """
    # This Xpath supports both Main_Queries-type files and batch query files
    xpathPattern = '/queries//queryGroup/*[@title="{title}"]|/queries/aQuery/*[@title="{title}"]'

    patterns = (None, '_', '-', '[-_]')

    for pattern in patterns:
        altTitle = re.sub(pattern, ' ', title) if pattern else title
        xpath = xpathPattern.format(title=altTitle)
        elts = tree.xpath(xpath)  # returns empty list or list of elements found
        if len(elts) != 0:
            return elts

    return None

def _findOrCreateQueryFile(title, queryPath, regions, outputDir=None, tmpFiles=True,
                           regionMap=None, rewriteSetList=None, rewriteParser=None,
                           delete=True):
    '''
    Find a query with the given title either as a file (with .xml extension) or
    within an XML query file by searching queryPath. If the query with "title" is
    found in an XML query file, extract it to generate a batch query file and
    apply it to the given regions. If outputDir is given, files are written there
    rather than creating temp files that would be deleted when the program exits.
    '''
    sep = os.path.pathsep           # ';' on Windows, ':' on Unix
    items = queryPath.split(sep)

    parser = ET.XMLParser(remove_blank_text=True)

    for item in items:
        if os.path.isdir(item):
            pathname = pathjoin(item, title + '.xml')
            if os.path.isfile(pathname):
                return pathname
            else:
                continue

        # Find the query within an XML query file
        tree = ET.parse(item, parser=parser)
        elts = _findQueryByName(tree, title)

        if elts is None or len(elts) == 0:
            continue # to next item in QueryPath

        _logger.debug("Found query '{}' in {}".format(title, item))

        root = ET.Element("queries")
        aQuery = ET.Element("aQuery")
        root.append(aQuery)
        for region in regions:
            aQuery.append(ET.Element('region', name=region))

        queryElt = elts[0]
        aQuery.append(queryElt)

        if regionMap or rewriteSetList:
            # if a rewrite list element exists already, use it, otherwise create it.
            subtree = ET.ElementTree(element=queryElt)
            found = subtree.xpath('//labelRewriteList')
            if len(found) == 1:
                rewriteList = found[0]
            else:
                # create and insert a labelRewriteList element
                rewriteList = ET.Element('labelRewriteList', attrib={'append-values': 'true'})
                queryElt.append(rewriteList)

            if regionMap:
                _addRegionMap(regionMap, rewriteList)

            if rewriteSetList:
                _addRewriteSet(rewriteSetList, rewriteParser, rewriteList, title)

        # Extract the query into a file to submit to ModelInterface
        if tmpFiles:
            path = getTempFile(suffix='.query.xml', delete=delete)
        else:
            outputDir = outputDir or getParam('GCAM.OutputDir')
            queryDir = pathjoin(outputDir, 'queries')
            mkdirs(queryDir)
            path = pathjoin(queryDir, title + '.xml')

        _logger.debug("Writing extracted query for '%s' to '%s'", title, path)
        tree = ET.ElementTree(root)
        tree.write(path, xml_declaration=True, encoding="UTF-8", pretty_print=True)
        return path


BatchQueryTemplate = """<?xml version="1.0"?>
<!-- WARNING: this file is automatically generated. Manual edits will be overwritten. -->
<ModelInterfaceBatch>
    <class name="ModelInterface.ModelGUI2.DbViewer">
        <command name="XMLDB Batch File">
            <scenario name="{scenario}"/>
            <queryFile>{queryFile}</queryFile>
            <outFile>{csvFile}</outFile>
            <xmldbLocation>{xmldb}</xmldbLocation>
            <batchQueryResultsInDifferentSheets>false</batchQueryResultsInDifferentSheets>
            <batchQueryIncludeCharts>false</batchQueryIncludeCharts>
            <batchQuerySplitRunsInDifferentSheets>false</batchQuerySplitRunsInDifferentSheets>
            <batchQueryReplaceResults>true</batchQueryReplaceResults>
        </command>
    </class>
</ModelInterfaceBatch>
"""

MultiCommandTemplate = """<?xml version="1.0"?>
<!-- WARNING: this file is automatically generated. Manual edits will be overwritten. -->
<ModelInterfaceBatch>
    <class name="ModelInterface.ModelGUI2.DbViewer">
        {batchCommands}
    </class>
</ModelInterfaceBatch>
"""

BatchCommandElement = """
        <command name="XMLDB Batch File">
            <scenario name="{scenario}"/>
            <queryFile>{queryFile}</queryFile>
            <outFile>{csvFile}</outFile>
            <xmldbLocation>{xmldb}</xmldbLocation>
            <batchQueryResultsInDifferentSheets>false</batchQueryResultsInDifferentSheets>
            <batchQueryIncludeCharts>false</batchQueryIncludeCharts>
            <batchQuerySplitRunsInDifferentSheets>false</batchQuerySplitRunsInDifferentSheets>
            <batchQueryReplaceResults>true</batchQueryReplaceResults>
        </command>
"""

#
# TBD: use new (in 4.3) -l flag to MI to route output to log file.
#
def _createJavaCommand(batchFile, miLogFile):
    # javaLibPathArg = '-Djava.library.path="%s"' % javaLibPath if javaLibPath else ""
    # command = 'java %s %s -jar "%s" -b "%s" %s' % (javaArgs, javaLibPathArg, jarFile, batchFile, redirect)

    # e.g., java %(GCAM.MI.JavaArgs)s -cp %(GCAM.MI.ClassPath)s ModelInterface/InterfaceMain -b "{batchFile}"
    template = getParam('GCAM.MI.BatchCommand')
    command = template.format(batchFile=batchFile)
    if miLogFile:
        command += ">> %s 2>&1" % miLogFile

    return command

def _copyToLogFile(logFile, filename, msg=''):
    with open(logFile, 'a') as m:
        with open(filename, 'r') as f:
            m.write(msg)
            map(m.write, f.readlines())


def _createBatchCommandElement(scenario, queryName, queryPath, outputDir=None, tmpFiles=True,
                               xmldb='', csvFile=None, regions=None, regionMap=None,
                               rewriters=None, rewriteParser=None, noDelete=False, saveAs=None):
    """
    Generate a <command> element for use in a multi-query batch file. The indicated
    query will be copied into a temporary file that is referenced by this <command>
    element.

    :param scenario: (str) the name of the scenario to perform the query on
    :param queryName: (str) the name of a query to execute
    :param queryPath: (str) a list of directories or XML filenames, separated
        by a colon (on Unix) or a semi-colon (on Windows)
    :param outputDir: (str) the directory in which to write the .CSV
        with query results. Defaults to value of GCAM.OutputDir.
    :param xmldb: (str) the pathname to the XML database to query, or '' to
        use in-memory DB
    :param csvFile: if None, query results are written to a computed filename.
    :param regions: (iterable of str) the regions you want to include in the query
    :param regionMap: (dict-like) keys are the names of regions that should be rewritten.
        The value is the name of the aggregate region to map into.
    :param rewriters: (list of tuples of (mapName, level)) list of mapping rewrites to
        apply to the query results, based on rewriteSets.xml. If level is specified, it
        overrides the level given in the mapping as defined in rewriteSets.xml.
    :param rewriteParser: (RewriteSetParser instance) parsed representation of
        rewriteSets.xml
    :param noDelete: (bool) if True, temporary files created by this function are
        not deleted (use for debugging)
    :param saveAs (str): alternative name to use to save the query results as
    :return: (str) the generated batch command string
    """
    basename = os.path.basename(queryName)
    mainPart, extension = os.path.splitext(basename)   # strip extension, if any

    # set default here so sphinx doc doesn't list all 32 regions
    regions = regions or GCAM_32_REGIONS

    delete = not noDelete

    # Look for both the literal name as given as well as the name with "-" and "_" replaced with " "
    queryFile = _findOrCreateQueryFile(basename, queryPath, regions, regionMap=regionMap,
                                       rewriteSetList=rewriters, rewriteParser=rewriteParser,
                                       outputDir=outputDir, tmpFiles=tmpFiles, delete=delete)

    if not queryFile:
        raise PygcamException("_createBatchCommand: file for query '%s' was not found in '%s'." % \
                              (basename, queryPath))

    if not csvFile:
        csvFile = "%s-%s.csv" % (saveAs or mainPart, scenario)    # compute default filename
        csvFile = csvFile.replace(' ', '_')                       # eliminate spaces for convenience

    outputDir = outputDir or getParam('GCAM.OutputDir')
    mkdirs(outputDir)
    csvPath = os.path.abspath(pathjoin(outputDir, csvFile))

    batchCommand = BatchCommandElement.format(scenario=scenario, queryFile=queryFile,
                                              csvFile=csvPath, xmldb=xmldb)

    return batchCommand


def createBatchFile(scenario, queries, xmldb='', queryPath=None, outputDir=None,
                    regions=None, regionMap=None, rewriteParser=None,
                    batchFileIn=None, batchFileOut=None,
                    tmpFiles=True, noDelete=False):
    """
    Create an optionally-temporary XML file that will run multiple queries, by extracting
    queries into separate temp files and referencing them from the batch query file.

    :param scenario: (str) the name of the scenario to perform the query on
    :param queries: (list of str query names and/or Query instances)
    :param xmldb: (str) path to XMLDB, or '' to use in-memory DB
    :param queryPath: (str) a list of directories or XML filenames, separated
        by a colon (on Unix) or a semi-colon (on Windows)
    :param outputDir: (str) the directory in which to write the .CSV
        with query results, default is value of GCAM.OutputDir.
    :param regions: (iterable of str) the regions you want to include in the query
    :param regionMap: (dict-like) keys are the names of regions that should be rewritten.
        The value is the name of the aggregate region to map into.
    :param rewriteParser: (RewriteSetParser instance) parsed representation of
        rewriteSets.xml
    :param batchFileIn: (str) the name of a pre-formed batch file to run
    :param batchFileOut: (str) where to write output from batchFileIn, if given
    :param tmpFiles: (bool) if True temporary files are used and deleted when the
        program exits, otherwise normal files are create in outputDir.
    :param noDelete: (bool) if True, temporary files created by this function are
        not deleted (use for debugging)
    :return: (str) the pathname of the temporary batch query file
    """
    from queryFile import Query

    commands = []

    for obj in queries:
        # handle both Query instances and simple query name strings
        if isinstance(obj, Query):
            queryName = obj.name
            rewriters = obj.rewriters
            saveAs    = obj.saveAs

            if not rewriteParser:
                raise PygcamException("No rewriteParser defined. Pass filename as argument to query sub-command or define GCAM.RewriteSetsFile")
        else:
            # Deprecated? Old style, simple list of query names.
            queryName = obj
            rewriters = None
            saveAs    = None

        queryName = queryName.strip()

        # N.B. A side-effect of this is that QueryFile elements can be commented out
        # by inserting a "#" at the start of the name, e.g., <query name="#foo" ...>
        if not queryName or queryName[0] == '#':    # ignore blank lines and comments
            continue

        # Extracts the named query into a temp file and returns XML text referencing the file.
        command = _createBatchCommandElement(scenario, queryName, queryPath, outputDir=outputDir,
                                             xmldb=xmldb, regions=regions, regionMap=regionMap,
                                             rewriters=rewriters, rewriteParser=rewriteParser,
                                             tmpFiles=tmpFiles, noDelete=noDelete, saveAs=saveAs)
        commands.append(command)

    # Add command to run pre-formed batch file, if given
    if batchFileIn:
        command = BatchCommandElement.format(scenario=scenario, queryFile=batchFileIn,
                                             csvFile=batchFileOut, xmldb=xmldb)
        commands.append(command)

    # Create the file batch-query.xml in the same dir as the CSV files. It can't be
    # a temp file because this step runs separately from the step running GCAM, and
    # the batch file would be either deleted prematurely or not at all.
    outputDir = outputDir or getParam('GCAM.OutputDir')
    if tmpFiles:
        batchFile = getTempFile(suffix='.batch.xml', delete=not noDelete, text=True)
    else:
        queryDir = pathjoin(outputDir, 'queries')
        mkdirs(queryDir)
        batchFile = pathjoin(queryDir, 'generated-batch-query.xml')

    batchCommands = ''.join(commands)
    contents = MultiCommandTemplate.format(batchCommands=batchCommands)

    saveToFile(contents, filename=batchFile)
    return batchFile


def runMultiQueryBatch(scenario, queries, xmldb='', queryPath=None, outputDir=None,
                       miLogFile=None, regions=None, regionMap=None, rewriteParser=None,
                       batchFileIn=None, batchFileOut=None, noRun=False, noDelete=False):
    """
    Create a single GCAM XML batch file that runs multiple queries, placing the
    each query's results in a file named of the form {queryName}-{scenario}.csv.

    :param scenario: (str) the name of the scenario to perform the query on
    :param queries: (list of str query names and/or Query instances)
    :param xmldb: (str) path to XMLDB, or '' to use in-memory DB
    :param queryPath: (str) a list of directories or XML filenames, separated
        by a colon (on Unix) or a semi-colon (on Windows)
    :param outputDir: (str) the directory in which to write the .CSV
        with query results, default is value of GCAM.OutputDir.
    :param regions: (iterable of str) the regions you want to include in the query
    :param regionMap: (dict-like) keys are the names of regions that should be rewritten.
        The value is the name of the aggregate region to map into.
    :param rewriteParser: (RewriteSetParser instance) parsed representation of
        rewriteSets.xml
    :param batchFileIn: (str) the name of a pre-formed batch file to run
    :param batchFileOut: (str) where to write output from batchFileIn, if given
    :param noRun: (bool) if True, print the command that would be executed, but
        don't run it.
    :param noDelete: (bool) if True, temporary files created by this function are
        not deleted (use for debugging)
    :return: none
    """
    batchFile = createBatchFile(scenario, queries, xmldb=xmldb, queryPath=queryPath,
                                outputDir=outputDir, regions=regions, regionMap=regionMap,
                                rewriteParser=rewriteParser, noDelete=noDelete,
                                batchFileIn=batchFileIn, batchFileOut=batchFileOut)

    runModelInterface(scenario, outputDir, xmldb=xmldb, batchFile=batchFile,
                      miLogFile=miLogFile, noDelete=noDelete, noRun=noRun)


# TBD: Test queryText and asDataFrame.
def runModelInterface(scenario, outputDir, csvFile=None, batchFile=None,
                      queryFile=None, queryText=None,  xmldb='',
                      miLogFile=None, noDelete=False, noRun=False,
                      asDataFrame=False):
    """
    Run a query file on the XML database given by `xmldb`, saving results
    in a CSV file.

    :param scenario: (str) the name of the scenario to perform the query on
    :param outputDir: (str) the directory in which to write the CSV
        with query results
    :param csvFile: (str) the file to create; ignored if batchFile is not None.
    :param batchFile: (str) the path to an existing batchFile; if not given,
        a temporary one will be created based on the other arguments. In no
        batchFile is specified, a csvFile must be provided.
    :param queryFile: (str) the path to the XML file holding the queries, or None
        if the batchFile is not None, in which case the queryFile should already
        be referenced in the batchFile.
    :param queryText: (str) the text of an XML query. Used only if queryFile and
        batchFile are both None, in which case the queryText is written to a tmp
        file that is used as the queryFile.
    :param xmldb: (str) the path to the XML database or '' to use in-memory DB
    :param miLogFile: (str) optional name of a log file to write ModelInterface output to.
        The value is the name of the aggregate region to map into.
    :param noRun: (bool) if True, the command is printed but not executed
    :param noDelete: (bool) if True, temporary files created by this function are
        not deleted (use for debugging)
    :param asDataFrame: (bool) see return, below
    :return: (str or DataFrame) if asDataFrame is True, return the result as a
        pandas DataFrame, if the query was successful, else None. If asDataFrame
        is False, return the absolute path to the generated CSV file if one was
        specified, else None.
    """
    filesToDelete = []   # so we can remove them when done

    def ignoreVar(name, using='batchFile'):
        _logger.warn('runModelInterface called with both %s and %s; the latter is ignored.', using, name)
        return None # just a convenience to set ignored variable to None

    if batchFile:
        if queryFile:
            queryFile = ignoreVar('queryFile')

        if queryText:
            queryText = ignoreVar('queryText')

        if csvFile:
            csvFile = ignoreVar('csvFile')

    elif queryFile and queryText:
        queryText = ignoreVar('queryText', using='queryFile')

    delete = not noDelete

    try:    # ensure that we delete temp files if needed
        if queryText:
            queryFile = getTempFile(suffix='.query.xml', delete=delete)
            filesToDelete.append(queryFile)
            _logger.debug("Creating temporary query file '%s'", queryFile)

        if csvFile and not batchFile:
            csvPath = os.path.abspath(pathjoin(outputDir, csvFile))
            _logger.debug("Writing results to: %s", csvPath)
        else:
            csvPath = None

        # TBD: test this. Might fail if not batchFile and not csvFile
        if not batchFile:
            # Create a batch file for ModelInterface to invoke the query on the named
            # scenario, and save the output where specified
            batchFile = getTempFile(suffix='.batch.xml', delete=delete)
            filesToDelete.append(batchFile)
            _logger.debug("Creating temporary batch file '%s'", batchFile)

            batchFileText = BatchQueryTemplate.format(scenario=scenario, queryFile=queryFile,
                                                      csvFile=csvPath, xmldb=xmldb)
            saveToFile(batchFileText, filename=batchFile)

        if miLogFile:
            mkdirs(os.path.dirname(miLogFile))
            if queryFile:
                _copyToLogFile(miLogFile, queryFile, "Query file: '%s'\n\n" % queryFile)
            _copyToLogFile(miLogFile, batchFile, "Batch file: '%s'\n\n" % batchFile)

        command = _createJavaCommand(batchFile, miLogFile)

        failed = False

        if noRun:
            print(command)
        else:
            _logger.debug(command)

        if getParamAsBoolean('GCAM.MI.UseVirtualBuffer'):   # deprecated as of GCAM 4.3
            with Xvfb():
                subprocess.call(command, shell=True)
        else:
            subprocess.call(command, shell=True)

        # The java program always exits with 0 status, but when the query fails,
        # it writes an error message to the CSV file. If this occurs, we delete
        # the file.
        if csvPath:
            try:
                with open(csvPath, 'r') as f:
                    line = f.readline()

                if re.search('java.*Exception', line, flags=re.IGNORECASE):
                    failed = True

            except Exception:
                failed = True

            finally:
                if failed:
                    _logger.error("Batch file '%s' failed. Deleting '%s'", queryFile, csvPath)
                    deleteFile(csvPath)
    except:
        raise

    finally:
        for filename in filesToDelete:
            TempFile.remove(filename, raiseError=False)

    if failed:
        return None

    if asDataFrame and csvPath:
        return readCsv(csvPath)

    return csvPath  # might be None if batchFile was provided; user must read file manually

def runBatchQuery(scenario, queryName, queryPath, outputDir, xmldb='',
                  csvFile=None, miLogFile=None, regions=None, regionMap=None,
                  rewriters=None, rewriteParser=None, noRun=False, noDelete=False,
                  saveAs=None):
    """
    Run a single query against GCAM's XML database given by `xmldb` (or
    computed from other parameters), saving the results into a CSV file.

    :param scenario: (str) the name of the scenario to perform the query on
    :param queryName: (str) the name of a query to execute
    :param queryPath: (str) a list of directories or XML filenames, separated
        by a colon (on Unix) or a semi-colon (on Windows)
    :param outputDir: (str) the directory in which to write the .CSV
        with query results
    :param xmldb: (str) the path to the XML database, or '' to use in-memory DB
    :param csvFile: if None, query results are written to a computed filename.
    :param miLogFile: (str) optional name of a log file to write ModelInterface output to.
    :param regions: (iterable of str) the regions you want to include in the query
    :param regionMap: (dict-like) keys are the names of regions that should be rewritten.
        The value is the name of the aggregate region to map into.
    :param rewriters: (list of tuples of (mapName, level)) list of mapping rewrites to
        apply to the query results, based on rewriteSets.xml. If level is specified, it
        overrides the level given in the mapping as defined in rewriteSets.xml.
    :param rewriteParser: (RewriteSetParser instance) parsed representation of
        rewriteSets.xml
    :param noRun: (bool) if True, the command is printed but not executed
    :param noDelete: (bool) if True, temporary files created by this function are
        not deleted (use for debugging)
    :param saveAs (str): alternative name to use to save the query results as
    :return: (str) the absolute path to the generated .CSV file, or None
    """
    basename = os.path.basename(queryName)
    mainPart, extension = os.path.splitext(basename)   # strip extension, if any

    regions = regions or GCAM_32_REGIONS # set default here so it doesn't mess up doc for this method

    delete = not noDelete

    # Look for both the literal name as given as well as the name with "-" and "_" replaced with " "
    filename = _findOrCreateQueryFile(basename, queryPath, regions, regionMap=regionMap,
                                      rewriteSetList=rewriters, rewriteParser=rewriteParser,
                                      delete=delete)
    if not filename:
        raise PygcamException("runBatchQuery: file for query '%s' was not found." % basename)

    if not csvFile:
        csvFile = "%s-%s.csv" % (saveAs or mainPart, scenario)    # compute default filename
        csvFile = csvFile.replace(' ', '_')                       # eliminate spaces for convenience

    csvPath = runModelInterface(scenario, filename, outputDir, csvFile, xmldb=xmldb,
                                miLogFile=miLogFile, noDelete=noDelete, noRun=noRun)
    return csvPath

# Deprecated: there no need for this once multiple queries is working
def _runSingleQueryBatch(scenario, xmldb='', queryNames=None, queryNodes=None, queryPath=None,
                        outputDir=None, rewriteParser=None, miLogFile=None, regions=None,
                        regionMap=None, noRun=False, noDelete=False):
    queryNames = queryNames or []
    queryNodes = queryNodes or []

    for queryName in queryNames:
        queryName = queryName.strip()

        if not queryName or queryName[0] == '#':    # allow blank lines and comments
            continue

        if queryName == 'exit':
            _logger.warn('Found "exit"; exiting batch query processing')
            return

        _logger.info("Processing query '%s'", queryName)

        runBatchQuery(scenario, queryName, queryPath, outputDir, xmldb=xmldb,
                      miLogFile=miLogFile, regions=regions, regionMap=regionMap,
                      noRun=noRun, noDelete=noDelete)

    for queryNode in queryNodes:
        queryName = queryNode.name
        saveAs    = queryNode.saveAs
        rewriters = queryNode.rewriters

        runBatchQuery(scenario, queryName, queryPath, outputDir, xmldb=xmldb,
                      miLogFile=miLogFile, regions=regions, regionMap=regionMap,
                      rewriters=rewriters, rewriteParser=rewriteParser,
                      noRun=noRun, noDelete=noDelete, saveAs=saveAs)

def sumYears(files, skiprows=1, interpolate=False):
    """
    For each file given, sum all values in each year column and create
    a file holding the result. Each resulting filename has the same basename
    as the input file but ending with '-sum.csv'.

    :param files: (list of str) Filenames to process
    :param skiprows: (int) the number of rows to skip prior to column headers
    :param interpolate: (bool) if True, interpolate annual values between time-steps
    :return: none
    """
    csvFiles = map(ensureCSV, files)
    dframes  = map(lambda fname: readCsv(fname, skiprows=skiprows, interpolate=interpolate), csvFiles)

    # TBD: preserve columns that have a single value only? Maybe this collapses into sumYearsByGroup()?
    for df, fname in zip(dframes, csvFiles):
        root, ext = os.path.splitext(fname)
        outFile = root + '-sum' + ext
        yearCols = filter(str.isdigit, df.columns)

        with open(outFile, 'w') as f:
            sums = df[yearCols].sum()
            csvText = sums.to_csv(None)
            f.write("%s\n%s\n" % (outFile, csvText))

# TBD: pass an output directory?
def sumYearsByGroup(groupCol, files, skiprows=1, interpolate=False):
    """
    Group data for each time-step (or interpolated annual values) by the given
    column (with categorical data like region or sector), and sum all
    members of each group to produce a time-series for each group. Equivalent to
    a SQL "group by" operation. For each input file, a new .CSV file is written
    with the name formed by the basename of the original file, followed by
    "-groupby-" and the groupCol. For example, given the file "foobar.csv" and
    groupCol "region", the file "foobar-groupby-region.csv" would be generated.
    Tests that all rows have the same units; otherwise raises an error.

    :param groupCol: (str) the column with categorical data to group by.
    :param files: (list of str) Filenames to process
    :param skiprows: (int) the number of rows to skip prior to column headers
    :param interpolate: (bool) if True, interpolate annual values between time-steps
    :return: none
    :raises CommandLineError: if the rows in the input file don't all have the same units
    """
    import numpy as np

    csvFiles = map(ensureCSV, files)
    dframes  = map(lambda fname: readCsv(fname, skiprows=skiprows, interpolate=interpolate), csvFiles)

    for df, fname in zip(dframes, csvFiles):
        units = df['Units'].unique()
        if len(units) != 1:
            raise CommandlineError("Can't sum results; rows have different units: %s" % units)

        root, ext = os.path.splitext(fname)
        name = groupCol.replace(' ', '_')     # eliminate spaces for general convenience
        outFile = '%s-groupby-%s%s' % (root, name, ext)

        cols = [groupCol] + filter(str.isdigit, df.columns)
        grouped = df[cols].groupby(groupCol)
        df2 = grouped.aggregate(np.sum)
        df2['Units'] = units[0]         # add these units to all rows

        with open(outFile, 'w') as f:
            csvText = df2.to_csv(None)
            label = outFile
            f.write("%s\n%s\n" % (label, csvText))

def csv2xlsx(inFiles, outFile, skiprows=0, interpolate=False, years=None, startYear=0):
    """
    Convert a set of CSV files representing GCAM query results into an XLSX file
    with an index page linked by the file names to the sheets with the results.

    :param inFiles: (list of str) the names of CSV files to read.
    :param outFile: (str) the name of the XLSX file to create
    :param skiprows: (int) the number of rows to skip in the CSV files before processing.
    :param interpolate: (bool) if True, interpolate annual values between time-steps.
    :param years: (str) the years to extract from the CSV files; must be of the form
      XXXX-YYYY, e.g. 2010-2050.
    :param startYear: (int) If interpolating, the year to begin interpolation
    :return: none
    """
    import pandas as pd

    csvFiles = map(ensureCSV, inFiles)

    try:
        dframes  = map(lambda fname: readCsv(fname, skiprows=skiprows, interpolate=interpolate,
                                             years=years, startYear=startYear), csvFiles)
    except Exception as e:
        raise CommandlineError("readCsv failed: %s" % e)

    formatStr = getParam('GCAM.ExcelNumberFormat')

    basenames = map(os.path.basename, csvFiles)
    sheetNum = 1
    outFile = ensureExtension(outFile, '.xlsx')
    with pd.ExcelWriter(outFile, engine='xlsxwriter') as writer:
        workbook = writer.book
        numFormat = workbook.add_format({'num_format': formatStr}) if formatStr else None
        linkFmt   = workbook.add_format({'font_color': 'blue', 'underline': True})

        # Create an index sheet
        indexSheet = workbook.add_worksheet('index')
        indexSheet.write_string(0, 1, 'Links to query results')
        maxlen = max(map(len, basenames))
        indexSheet.set_column('B:B', maxlen)

        for i, name in enumerate(basenames):
            row = i+1
            indexSheet.write(row, 0, row)
            indexSheet.write_url(row, 1, "internal:%d!A1" % row, linkFmt, name)

        for df, fname in zip(dframes, basenames):
            sheetName = str(sheetNum)
            sheetNum += 1
            dropExtraCols(df, inplace=True)
            df.to_excel(writer, index=None, sheet_name=sheetName, startrow=3, startcol=0)
            worksheet = writer.sheets[sheetName]

            if numFormat:
                # get the numerical indices of all column names that are numeric (i.e., years)
                yearIndices = map(lambda pair: pair[0], filter(lambda pair: str.isdigit(pair[1]), enumerate(df.columns)))
                for idx in yearIndices:
                    worksheet.set_column(idx, idx, None, numFormat)

            worksheet.write_string(0, 0, "Filename:")
            worksheet.write_string(0, 1, fname)
            worksheet.write_url(1, 0, "internal:index!A1", linkFmt, "Back to index")


def queryMain(args):
    # """
    # Main driver for query sub-command
    #
    # :param args:
    # :return: none
    # """
    miLogFile   = getParam('GCAM.MI.LogFile')
    outputDir   = args.outputDir or getParam('GCAM.OutputDir')
    groupDir    = args.groupDir
    scenario    = args.scenario
    sandbox     = args.workspace or pathjoin(getParam('GCAM.SandboxDir'), groupDir, scenario)
    xmldb       = args.xmldb     or pathjoin(sandbox, 'output', getParam('GCAM.DbFile'))
    queryPath   = args.queryPath or getParam('GCAM.QueryPath')
    queryFile   = args.queryXmlFile
    regionFile  = args.regionMap or getParam('GCAM.RegionMapFile')
    regions     = args.regions.split(',') if args.regions else GCAM_32_REGIONS
    queryNames  = args.queryName
    noDelete    = args.noDelete
    prequery    = args.prequery
    versionNum  = getParam('GCAM.VersionNumber')
    versionInfo = parse_version_info()
    inMemory        = versionInfo > (4, 2) and getParamAsBoolean('GCAM.InMemoryDatabase')
    internalQueries = versionInfo > (4, 2) and (inMemory or getParamAsBoolean('GCAM.RunQueriesInGCAM'))
    batchMultiple   = internalQueries or getParamAsBoolean('GCAM.BatchMultipleQueries')
    rewriteSetsFile = args.rewriteSetsFile or getParam('GCAM.RewriteSetsFile')
    batchFileIn  = args.batchFile
    batchFileOut = os.path.abspath(pathjoin(outputDir, args.batchOutput))

    # Post-GCAM queries are not possible when using in-memory database.
    # The 'prequery' step writes the XMLDBDriver.properties file used
    # by GCAM to query the in-memory database before exiting.
    if prequery and versionInfo <= (4, 2):
        _logger.info('Skipping pre-query step for gcam-v%s', versionNum)
        return

    if internalQueries and not prequery:
        _logger.info('Skipping post-GCAM query step: GCAM runs queries internally')
        return

    if inMemory and not prequery:
        _logger.info('Skipping post-GCAM query step: using in-memory database')
        return

    if not (xmldb or inMemory):
        raise CommandlineError('Must specify xmldb if not using in-memory database')

    if queryNames:
        _logger.debug("Query names: %s", queryNames)

    queryFileObj = QueryFile.parse(queryFile) if queryFile else None
    queryNodes = queryFileObj.queries if queryFileObj else []
    queries = queryNames + queryNodes

    if not (queryNames or queryFileObj or batchFileIn):
        raise CommandlineError("Error: At least one query name, query XML, or batch file must be specified")

    rewriteParser = RewriteSetParser.parse(rewriteSetsFile) if rewriteSetsFile else None
    regionMap = readRegionMap(regionFile) if regionFile else None

    mkdirs(outputDir)

    # TBD: if setupWorkspace is performed in 'setup' step, can be removed from here
    # When writing the XMLDB to disk, we call ModelInterface separately. When
    # using an in-memory database, GCAM runs the queries for us, so here we
    # just create the XMLDBDriver.properties and batch files and return.
    if args.prequery:
        exeDir = getExeDir(sandbox, chdir=True)
        batchFile = createBatchFile(scenario, queries, queryPath=queryPath, outputDir=outputDir,
                                    regions=regions, regionMap=regionMap, rewriteParser=rewriteParser,
                                    batchFileIn=batchFileIn, batchFileOut=batchFileOut,
                                    tmpFiles=False, noDelete=noDelete) \
            if batchMultiple else ''

        filterFile = getParam('GCAM.FilterFile')
        writeXmldbDriverProperties(inMemory=inMemory, outputDir=exeDir, filterFile=filterFile,
                                   batchFile=batchFile, batchLog='logs/batch-query.log')
        return

    xmldb = os.path.abspath(xmldb)

    if miLogFile:
        miLogFile = os.path.abspath(pathjoin(outputDir, miLogFile))
        deleteFile(miLogFile)       # remove it, if any, to start fresh

    # If not a prequery step, we're running queries post-GCAM, which means a database on disk
    # For now, we support running multiple queries in a single batch file, or the old way,
    # running each one individually. The latter is probably not needed, except for debugging.
    if batchMultiple:
        runMultiQueryBatch(scenario, queries, xmldb=xmldb, queryPath=queryPath, outputDir=outputDir,
                           miLogFile=miLogFile, regions=regions, regionMap=regionMap,
                           batchFileIn=batchFileIn, batchFileOut=batchFileOut,
                           rewriteParser=rewriteParser, noRun=args.noRun, noDelete=noDelete)
    else:
        # (Deprecated) Otherwise run them individually.
        _runSingleQueryBatch(scenario, xmldb=xmldb, queryNames=queryNames, queryNodes=queryNodes,
                             queryPath=queryPath, outputDir=outputDir, rewriteParser=rewriteParser,
                             miLogFile=miLogFile, regions=regions, regionMap=regionMap,
                             noRun=args.noRun, noDelete=args.noDelete)



#
# UNIT CONVERSIONS
#
# Pythonized from gcam-data-system/_common/assumptions/unit_conversions.R
#

'''
This module includes a large number of pre-defined symbolic constants to
support unit conversion. Values can be accessed, and additional values added,
programmatically via this API.

Use ``getUnits()`` to get the singleton instance of the ``UnitNamespace``
class, and set values prior to use in, say, the API to the :doc:`pygcam.chart`
module. For example,

  .. code-block:: python

     from pygcam.units import getUnits

     u = getUnits()
     u.NewConstant = 123.456
'''

from .error import PygcamException
from .log import getLogger
from .utils import coercible

_logger = getLogger(__name__)

def getUnits(other=None):
    """
    Returns a singleton instance of a private namespace-like class
    that can set and get unit conversions.

    :param other: (dict-like) an object supporting the items() method
        that returns pairs of (name, value) that are added to the built-in
        unit conversion look-up table. Note that elements in `other` will
        overwrite existing entries of the same name.
    :return: (UnitNamespace) returns a singleton instance, i.e., subsequent
        calls will return the same object, though additional names can be
        passed on each call.
    """
    units = UnitNamespace.getUnits()

    if other:
        for name, value in other.items():
            units.set(name, value)

    return units

class UnitNamespace(object):
    """
    UnitNamespace is a namespace-like class that stores the names of
    unit conversions, allowing these names to be used in the GCAM tool
    "chart" sub-command to specify values to multiply or divide by to
    convert units for plotting. Names can be set and accessed using
    '.' notation. That is, after calling ``u = getUnits()``, you
    can access a conversion named, for example, 'MJ_to_EJ' as
    ``u.MJ_to_EJ``, or set a new value for, say, 'foo' simply by
    setting ``u.foo = value``.

    Note that the getUnits() method should generally be used rather
    than calling UnitNamespace() directly to ensure that the single
    instance of the class is returned.
    """
    instance = None
    conversions = {}        # not an ivar since we hijack __getattr__

    @classmethod
    def getUnits(cls):
        if not cls.instance:
            cls.instance = cls()

        return cls.instance

    def __getattr__(self, name):
        return self.get(name, raiseError=True)

    def __setattr__(self, name, value):
        self.conversions[name] = value

    def get(self, name, raiseError=True):
        """
        Get the value of a variable defined in the unitConversion module.
        Basically an "eval" operation.

        :param name: (str) the name of a variable to look up.
        :param raiseError: (bool) if True, raise an error if converter is not found.
        :return: (float or None) the value of the named variable, or None if not found.
        """
        try:
            return self.conversions[name]

        except KeyError:
            if raiseError:
                raise PygcamException('Unknown unit conversion string "%s"' % name)
            return None

    def set(self, name, value):
        self.conversions[name] = value

    def convert(self, string, raiseError=True):
        """
        Convert the given `string` to its numerical value, either by
        direct type conversion, or if this fails, by look-up as the
        name of a defined unit conversion.

        :param string: (str) a string representing a float or the
            name of a defined unit conversion.
        :param raiseError: (bool) if True, an error is raised if the
            string is neither a number nor a defined unit conversion.
        :return: (float or None) returns the converted value, or if raiseError
            is False and the string is not a recognized unit conversion,
            None is returned.
        """
        value = coercible(string, float, raiseError=False)
        if value is not None:
            return value

        value = self.get(string, raiseError=raiseError)
        if value is not None:
            return value

        return None

    def __init__(self):
        """
        Initialize the unit conversion look-up table.
        """
        s = self    # just to make the following more readable

        # basic SI units and relationships
        s.k = 1e3
        s.M = 1e6
        s.G = 1e9
        s.T = 1e12
        s.P = 1e15
        s.E = 1e18
        s.k_to_ones = 1e3
        s.ones_to_k = 1e-3
        s.M_to_ones = 1e6
        s.ones_to_M = 1e-6
        s.M_to_k = 1e3
        s.k_to_M = 1e-3
        s.G_to_ones = 1e9
        s.ones_to_G = 1e-9
        s.G_to_k = 1e6
        s.k_to_G = 1e-6
        s.G_to_M = 1e3
        s.M_to_G = 1e-3
        s.P_to_M = 1e9
        s.M_to_P = 1e-9
        s.P_to_G = 1e6
        s.G_to_P = 1e-6
        s.P_to_T = 1e3
        s.T_to_P = 1e-3
        s.E_to_G = 1e9
        s.G_to_E = 1e-9

        s.E_to_M = s.E / s.M
        s.M_to_E = s.M / s.E

        # Carbon <--> CO2
        s.C_to_CO2 = 44./12
        s.CO2_to_C = 12./44

        # mass conversions
        s.kg_to_g =   s.k_to_ones
        s.g_to_kg =   s.ones_to_k
        s.Mg_to_kg =  s.M_to_k
        s.kg_to_Mg =  s.k_to_M
        s.Mg_to_ton = 1000./908
        s.ton_to_Mg = 0.908
        s.Mt_to_ton = s.Mg_to_ton
        s.ton_to_Mt = s.ton_to_Mg
        s.Pg_to_Tg =  s.P_to_T
        s.Tg_to_Pg =  s.T_to_P

        # calorie conversions
        s.Pcal_to_Mcal = s.P_to_M
        s.Mcal_to_Pcal = s.M_to_P
        s.Pcal_to_Gcal = s.P_to_G
        s.Gcal_to_Pcal = s.G_to_P

        # volume conversions
        s.m3_to_Gm3 =  1e-9
        s.m3_to_bm3 =  s.m3_to_Gm3
        s.km3_to_Mm3 = 1e3
        s.Mm3_to_km3 = 1e-3

        # area conversions
        s.km2_to_m2 = s.k * s.k
        s.ha_to_m2 = 1e4
        s.m2_to_ha = 1e-4
        s.km2_to_ha = s.km2_to_m2 / s.ha_to_m2
        s.ha_to_km2 = 1 / s.km2_to_ha
        s.kkm2_to_ha = 1000 * s.km2_to_ha
        s.thous_km2_to_ha = s.kkm2_to_ha    # alias,
        s.kkm2_to_Mha = s.kkm2_to_ha / s.M
        s.bm2_to_ha =  1e5
        s.ha_to_bm2 =  1e-5
        s.ha_to_m2 =   1e4
        s.m2_to_ha =   1e-4
        s.acre_to_m2 = 4046.85
        s.m2_to_acre = 1 / s.acre_to_m2
        s.Gm2_to_km2 = s.G_to_k
        s.bm2_to_km2 = s.G_to_k
        s.km2_to_Gm2 = s.k_to_G
        s.km2_to_bm2 = s.k_to_G
        s.bm2_to_m2 =  s.G_to_ones
        s.Gm2_to_m2 =  s.G_to_ones
        s.ft2_to_m2 =  0.0929
        s.Mft2_to_m2 = s.ft2_to_m2 * s.M_to_ones
        # million square feet to billion square meters
        s.Mft2_to_Gm2 = s.ft2_to_m2 * s.M_to_G
        s.Mft2_to_bm2 = s.Mft2_to_Gm2
        # square feet to billion square meters
        s.ft2_to_Gm2 = s.ft2_to_m2 * s.ones_to_G
        s.ft2_to_bm2 = s.ft2_to_Gm2

        # yield conversions
        s.Mg_per_ha_to_kg_per_m2 = s.M_to_k / s.ha_to_m2

        # time
        s.year_to_day = 365.25
        s.day_to_year = 1 / s.year_to_day
        s.day_to_hour = 24
        s.hour_to_day = 1. / s.day_to_hour
        s.year_to_hour = s.year_to_day * s.day_to_hour

        # energy unit conversions
        s.kbbl_to_bbl = s.k_to_ones
        s.bbl_to_tonne_RFO = 1 / 6.66
        s.bbl_to_tonne_distillate = 1 / 7.46
        s.tonne_to_GJ_distillate = 42.91
        s.tonne_to_GJ_RFO = 40.87

        s.MJ_to_EJ = s.M_to_E
        s.EJ_to_MJ = s.E_to_M
        s.GJ_to_EJ = s.G_to_E
        s.EJ_to_GJ = s.E_to_G
        s.kWh_to_MJ = 3.6
        s.MWh_to_GJ = s.kWh_to_MJ
        s.kWh_to_GJ = 0.0036
        s.kWh_to_GJ = s.kWh_to_MJ * s.M_to_G
        s.MWh_to_EJ = 3.6e-9
        s.GWh_to_EJ = 3.6e-6
        s.TWh_to_EJ = 3.6e-3
        s.btu_to_kJ = 1.0551
        s.kJ_to_btu = 1 / s.btu_to_kJ
        s.MJ_to_btu = s.kJ_to_btu * s.M_to_k
        s.btu_to_MJ = 1 / s.MJ_to_btu
        s.MMbtu_to_MJ = s.M_to_ones * s.btu_to_MJ
        s.MJ_to_MMbtu = 1 / s.MMbtu_to_MJ
        s.MJ_to_MMBtu = s.MJ_to_MMbtu
        s.MMBtu_to_MJ = s.MMbtu_to_MJ
        s.quad_to_EJ = s.btu_to_kJ
        s.EJ_to_quad = 1 / s.quad_to_EJ

        s.TOE_to_MJ = 41868
        s.MTOE_to_EJ = s.M * s.TOE_to_MJ * s.MJ_to_EJ
        s.EJ_to_MTOE = 1 / s.MTOE_to_EJ

        # trillion and thousand btu to EJ
        s.Tbtu_to_EJ = 0.0010551
        s.kbtu_to_EJ = 1.0551e-12

        # dollar inflation/deflation
        s.USD_1990_to_2005 = 1.383
        s.USD_1990_to_2007 = 1.470
        s.USD_1990_to_2010 = 1.510
        s.USD_1990_to_1975 = 0.4649
        s.USD_1996_to_1975 = 0.4049
        s.USD_1997_to_1975 = 0.3983
        s.USD_1998_to_1975 = 0.3939
        s.USD_1999_to_1975 = 0.3883
        s.USD_2000_to_1975 = 0.380
        s.USD_2001_to_1975 = 0.3711
        s.USD_2002_to_1975 = 0.3647
        s.USD_2003_to_1975 = 0.3571
        s.USD_2004_to_1975 = 0.3472
        s.USD_2005_to_1975 = 0.3362
        s.USD_2006_to_1975 = 0.3257
        s.USD_2007_to_1975 = 0.317
        s.USD_2008_to_1975 = 0.3104
        s.USD_2009_to_1975 = 0.3104



from abc import ABCMeta, abstractmethod
import json
import math
import os
import numpy as np
import pandas as pd

from SALib.analyze.delta import analyze as delta_analyzer
from SALib.analyze.dgsm import analyze as dgsm_analyzer

from SALib.analyze import fast, morris, sobol
from SALib.sample import saltelli, fast_sampler
from SALib.sample import morris as morris_sampler
from SALib.sample.latin import sample as latin_sampler
from SALib.sample.finite_diff import sample as finite_diff_sampler

from SALib.test_functions import Ishigami

# TBD: Save data compressed (opt), with zlib or gzip pkgs.

DFLT_PROBLEM_FILE = 'problem.csv'
DFLT_INPUTS_FILE  = 'inputs.csv'
DFLT_RESULTS_FILE = 'results.csv'
DFLT_GROUPS_FILE  = 'groups.csv'
DFLT_LINKED_FILE  = 'linkedCols.json'

# From https://waterprogramming.wordpress.com/2013/08/05/running-sobol-sensitivity-analysis-using-salib/
# If the confidence intervals of your dominant indices are larger than
# roughly 10% of the value itself, you may want to consider increasing
# your sample size as computation permits. For total-order indices to be
# important, they will usually need to be above 0.05 at the very least
# (the most dominant parameters will have values upward of 0.8).
#
# The cross-sampling scheme creates a total of 2N(p+1) total parameter
# sets to be run in your model, where N is the initial number of Sobol
# sequence values generated, and p is the number of parameters.

# Approach: use Spearman method to identify the top 5 or 6 parameters, then
# use Sobol analysis to zero in on them?? Maybe the Sobol approach is simply
# intractable with a model like GCAM? For 39 parameters, 2400 runs doesn't
# come close to adequate. Say we use a Sobol sequence of len 500, this would
# require 39,000 trials. (Of course the CB manual suggests 10,000 runs...)


class SAException(Exception):
    pass

class SensitivityAnalysis(object):
    """
    Abstract superclass for Sensitivity Analysis methods from SALib. Stores sets
    of method name and args, parameter descriptions, samples, and model results
    in a directory "package" with an ".sa" extension. Ensures that analysis
    methods are called with the same arguments used to produce the samples.
    """
    __metaclass__ = ABCMeta

    def __init__(self, pkgPath, problemFile=None, inputsFile=None, resultsFile=None):
        self.pkgPath   = pkgPath
        self.problem   = None
        self.inputs    = None
        self.inputsDF  = None
        self.results   = None
        self.resultsDF = None
        self.s2df      = None
        self.kwargs    = {}

        self.argsFile    = os.path.join(pkgPath, 'args.json')
        self.problemFile = os.path.join(pkgPath, problemFile or DFLT_PROBLEM_FILE)
        self.inputsFile  = os.path.join(pkgPath, inputsFile  or DFLT_INPUTS_FILE)
        self.resultsFile = os.path.join(pkgPath, resultsFile or DFLT_RESULTS_FILE)

        if os.path.lexists(self.problemFile):
            self.loadProblem()

    def loadProblem(self, filename=None, sep=','):
        self.problemFile = filename or self.problemFile
        data = pd.read_table(self.problemFile, sep=sep)

        self.problem = {'num_vars' : len(data),
                        'names'  : list(data.name),
                        'bounds' : data.as_matrix(columns=['low','high']),
                        #'groups' : None # required for Morris
                        }
        return self.problem

    def loadResults(self, resultName, resultsFile=None, sep=','):
        """
        Load model results from a CSV file.

        :param resultName: (str) the name of the model output
        :param resultsFile: (str) the path of the results file
        :param sep: (str) column separator
        :return: (pandas.DataFrame) contents of the results file
        """
        self.resultsFile = resultsFile or self.resultsFile
        self.resultsDF = pd.read_table(self.resultsFile, sep=sep)
        self.results = self.resultsDF[resultName].as_matrix()
        self.resultName = resultName
        return self.results

    def saveResults(self, results):
        self.results = results
        self.resultsDF = pd.DataFrame(data=results)
        self.resultsDF.to_csv(self.resultsFile, sep=',')

    def saveInputs(self, inputs=None, filename=None):
        inputs = inputs or self.inputs
        filename = filename or self.inputsFile
        varNames = self.problem['names']
        df = self.inputsDF = pd.DataFrame(data=inputs, columns=varNames)
        df['trialNum'] = df.index
        df.to_csv(filename, index=False, sep=',')

    def loadInputs(self, filename=None, sep=','):
        filename = filename or self.inputsFile
        if not filename:
            raise SAException("Can't loadInputs: filename is None")

        self.inputsDF = pd.read_table(filename, sep=sep, index_col='trialNum')
        self.inputs  = self.inputsDF.as_matrix()

    def saveArgs(self):
        with open(self.argsFile, 'w') as f:
            json.dump(self.kwargs, f)
            f.write('\n')

    def loadArgs(self):
        with open(self.argsFile) as f:
            self.kwargs = json.load(f)
        return self.kwargs

    def storeKwargs(*args, **kwargs):
        self = args[0]
        self.kwargs.update(kwargs)

    def getKwarg(self, key, kwargs):
        return kwargs.get(key, self.kwargs[key])

    def sample(*args, **kwargs):
        args = list(args)
        self = args.pop(0)
        self.trials = kwargs.get('trials', 1000)
        self.calcSecondOrder = kwargs.get('calc_second_order', True)
        self.N = kwargs.get('N', self.predictN(self.trials, self.calcSecondOrder))

        self.inputs = self._sample(*args, **kwargs)
        self.saveInputs()
        self.saveArgs()
        return self.inputs

    def analyze(*args, **kwargs):
        args = list(args)
        self = args.pop(0)
        self.loadInputs()
        self.loadArgs()

        # Handle nan values by replacing them with the mean of all other results
        Y = self.results
        nans = np.isnan(Y)
        if nans.any():
            Y[nans] = Y[~nans].mean()

        if kwargs.get('print_to_console', False):
            print("\n%s:" % self.__class__.__name__)

        analysisDict = self._analyze(*args, **kwargs)

        if 'S2' in analysisDict:
            S2 = analysisDict['S2']
            del analysisDict['S2']
            S2_conf = analysisDict['S2_conf']
            del analysisDict['S2_conf']

            names = self.problem['names']
            D = len(names)
            s2Dict = {'name1' : [], 'name2' : [], 'S2': [], 'S2_conf' : []}

            for j in range(D):
                name1 = names[j]
                for k in range(j + 1, D):
                    s2Dict['name1'].append(name1)
                    s2Dict['name2'].append(names[k])
                    s2Dict['S2'].append(S2[j, k])
                    s2Dict['S2_conf'].append(S2_conf[j, k])

            self.s2df = df = pd.DataFrame(data=s2Dict, columns=['name1', 'name2', 'S2', 'S2_conf'])
            #df.set_index(['name1', 'name2'], inplace=True)
            df.sort_values(by='S2', ascending=False, inplace=True)

        df = self.analysis = pd.DataFrame(data=analysisDict)
        df['names'] = self.problem['names']
        df['abs_S1'] = abs(df['S1'])
        df.set_index('names', inplace=True)
        df.sort_values(by='abs_S1', ascending=False, inplace=True)
        return df

    def predictN(self, trials, calcSecondOrder=False):
        '''
        Computes the value of N required to produce the given number
        of samples, per SA method.

        :param trials: (int) the number of total samples desired
        :param calcSecondOrder: (bool) whether to calculate second-order
           sensitivity indices (for Sobol method only).
        :return: (int) the value of N to use to produce `trials`
        '''
        nVars = self.problem['num_vars']
        return self._predictN(trials, nVars, calcSecondOrder)

    @abstractmethod
    def _predictN(self, trials, nVars, calcSecondOrder):
        pass

    @abstractmethod
    def _sample(self, **kwargs):
        """
        _sample() methods must set self.kwargs to hold all values that
        must be the same in _analyze(). These are saved to disk.
        """
        pass

    @abstractmethod
    def _analyze(self, **kwargs):
        pass

# TBD: Implement this
class MonteCarlo(SensitivityAnalysis):
    def __init__(self, pkgPath, problemFile=None, inputsFile=None, resultsFile=None):
        super(MonteCarlo, self).__init__(pkgPath, problemFile=problemFile,
                                         inputsFile=inputsFile, resultsFile=resultsFile)

    def _sample(*args, **kwargs):
        self = args[0]
        self.storeKwargs(**kwargs)
        return None # sample(self.problem, self.N, calc_second_order=self.calcSecondOrder)

    def _analyze(*args, **kwargs):
        '''
        Returns a dictionary with key 'S1', which is a list of size D (the
        number of parameters) containing the indices in the same order as in
        the parameter file. The values are the normalized rank correlations
        of each parameter to the designated output variable.
        :param args:
        :param kwargs:
        :return:
        '''
        self = args[0]
        print_to_console = kwargs.get('print_to_console', False)

        return None # analyze(self.problem, self.results, print_to_console=print_to_console)

    def _predictN(self, trials, nVars, calcSecondOrder):
        return trials

class Sobol(SensitivityAnalysis):
    """
    Provides an interface to SALib's Sobol Sensitivity sampling and analysis methods.
    """
    def __init__(self, pkgPath, problemFile=None, inputsFile=None, resultsFile=None):
        super(Sobol, self).__init__(pkgPath, problemFile=problemFile,
                                    inputsFile=inputsFile, resultsFile=resultsFile)

    def _sample(*args, **kwargs):
        self = args[0]
        self.storeKwargs(calc_second_order=self.calcSecondOrder)
        return saltelli.sample(self.problem, self.N, calc_second_order=self.calcSecondOrder)

    def _analyze(*args, **kwargs):
        self = args[0]

        calc_second_order = self.getKwarg('calc_second_order', kwargs)

        num_resamples     = kwargs.get('num_resamples', 100)
        conf_level        = kwargs.get('conf_level', 0.95)
        print_to_console  = kwargs.get('print_to_console', False)
        parallel          = kwargs.get('parallel', False)
        n_processors      = kwargs.get('n_processors', None)

        return sobol.analyze(self.problem, self.results, calc_second_order=calc_second_order,
                             num_resamples=num_resamples, conf_level=conf_level,
                             print_to_console=print_to_console, parallel=parallel,
                             n_processors=n_processors)

    def _predictN(self, trials, nVars, calcSecondOrder):
        multiple = 2 if calcSecondOrder else 1
        return int(math.ceil(trials / (multiple * nVars + 2)))


class FAST(SensitivityAnalysis):
    """
    Provides an interface to SALib's Fourier Amplitude Sensitivity Test (FAST)
    sampling and analysis methods.
    """
    def __init__(self, pkgPath, problemFile=None, inputsFile=None, resultsFile=None):
        super(FAST, self).__init__(pkgPath, problemFile=problemFile,
                                   inputsFile=inputsFile, resultsFile=resultsFile)

    def _sample(*args, **kwargs):
        self = args[0]

        N = self.N
        M = kwargs.get('M', 4)
        self.storeKwargs(N=N, M=M)

        return fast_sampler.sample(self.problem, N=N, M=M)

    def _analyze(*args, **kwargs):
        self = args[0]

        M = self.getKwarg('M', kwargs)
        print_to_console = kwargs.get('print_to_console', False)

        return fast.analyze(self.problem, self.results, M=M,
                            print_to_console=print_to_console)

    def _predictN(self, trials, nVars, calcSecondOrder):
        return int(math.ceil(trials / nVars))


class Morris(SensitivityAnalysis):
    """
    Provides an interface to SALib's Method of Morris sampling and analysis methods.
    """
    def __init__(self, pkgPath, problemFile=None, inputsFile=None, resultsFile=None,
                 groupsFile=None):
        super(Morris, self).__init__(pkgPath, problemFile=problemFile,
                                     inputsFile=inputsFile, resultsFile=resultsFile)

        # TBD: test 'groups' functionality
        self.groupsFile = self.groupsDF = None
        groupsFile = os.path.join(self.pkgPath, groupsFile or DFLT_GROUPS_FILE)

        if groupsFile and os.path.lexists(groupsFile):
            self.groupsFile = groupsFile
            self.groupsDF = pd.read_table(self.groupsFile)

        # set the 'groups' to None if no groups since sample method requires this
        if self.problem:
            self.problem['groups'] = self.groupsDF.as_matrix() if self.groupsDF else None

    # Maybe write kwargs to json file and reload this in analyze() to ensure same args used
    def _sample(*args, **kwargs):
        self = args[0]
        N = self.N
        num_levels           = kwargs.get('num_levels', 4)
        grid_jump            = kwargs.get('grid_jump', 2)
        optimal_trajectories = kwargs.get('optimal_trajectories', None)
        local_optimization   = kwargs.get('local_optimization', False)

        X = morris_sampler.sample(self.problem, N, num_levels, grid_jump,
                                  optimal_trajectories=optimal_trajectories,
                                  local_optimization=local_optimization)

        self.storeKwargs(num_levels=num_levels, grid_jump=grid_jump)
        return X

    def _analyze(*args, **kwargs):
        self = args[0]

        num_levels       = self.getKwarg('num_levels', kwargs)
        grid_jump        = self.getKwarg('grid_jump', kwargs)
        num_resamples    = kwargs.get('num_resamples', 1000)
        conf_level       = kwargs.get('conf_level', 0.95)
        print_to_console = kwargs.get('print_to_console', False)

        return morris.analyze(self.problem, self.inputs, self.results, num_resamples=num_resamples,
                              conf_level=conf_level, print_to_console=print_to_console,
                              grid_jump=grid_jump, num_levels=num_levels)

    def _predictN(self, trials, nVars, calcSecondOrder):
        return int(math.ceil(trials / (nVars + 1)))


class LatinSampler(SensitivityAnalysis):
    def __init__(self, pkgPath, analyzeFn=None, problemFile=None, inputsFile=None, resultsFile=None):
        super(LatinSampler, self).__init__(pkgPath, problemFile=problemFile,
                                     inputsFile=inputsFile, resultsFile=resultsFile)
        self.analyzeFn = analyzeFn
        self.defaultResamples = 10
        self.defaultConfLevel = 0.95
        self.defaultPrint = False

    def _sample(*args, **kwargs):
        self = args[0]
        N = self.N
        self.storeKwargs(N=N)
        X = latin_sampler(self.problem, N)
        return X

    def _analyze(*args, **kwargs):
        self = args[0]

        num_resamples    = kwargs.get('num_resamples', self.defaultResamples)
        conf_level       = kwargs.get('conf_level', self.defaultConfLevel)
        print_to_console = kwargs.get('print_to_console', self.defaultPrint)

        return self.analyzeFn(self.problem, self.inputs, self.results,
                              num_resamples=num_resamples,
                              conf_level=conf_level,
                              print_to_console=print_to_console)

class Delta(LatinSampler):
    """
    Provides an interface to SALib's Delta moment-independent measure 
    sampling and analysis methods.
    """
    def __init__(self, pkgPath, problemFile=None, inputsFile=None, resultsFile=None):
        super(Delta, self).__init__(pkgPath,
                                    analyzeFn=delta_analyzer, problemFile=problemFile,
                                    inputsFile=inputsFile, resultsFile=resultsFile)

    def _predictN(self, trials, nVars, calcSecondOrder):
        return trials


class DGSM(LatinSampler):
    """
    Provides an interface to SALib's Derivative-based Global Sensitivity 
    Measure (DGSM) sampling and analysis methods.
    """
    def __init__(self, pkgPath, problemFile=None, inputsFile=None, resultsFile=None):
        super(DGSM, self).__init__(pkgPath,
                                   analyzeFn=dgsm_analyzer, problemFile=problemFile,
                                   inputsFile=inputsFile, resultsFile=resultsFile)
        self.defaultResamples = 1000

    def _predictN(self, trials, nVars, calcSecondOrder):
        return int(math.ceil(trials / (nVars + 1)))

    def _sample(*args, **kwargs):
        self = args[0]
        N = self.N
        delta = kwargs.get('delta', 0.01)
        self.storeKwargs(N=N)
        X = finite_diff_sampler(self.problem, N, delta=delta)
        return X

if __name__ == "__main__":
    mcsTest = True

    if mcsTest:
        sa = Sobol('/Users/rjp/mcs/paper1/sims/s001/data.sa')
        sa.loadResults('corn-ci-all')
        df = sa.analyze(print_to_console=True)
        pd.set_option('display.width', 300)
        print(df)
        print(sa.s2df)

    else:
        for cls in (Sobol, FAST, Morris, Delta, DGSM):
            sa = cls('../tests/test.sa')

            kwargs = {'num_samples': 1000}

            if cls == Sobol:
                pass
            elif cls == FAST:
                kwargs['M'] = 4
            elif cls == Morris:
                kwargs['num_levels'] = 4
                kwargs['grid_jump'] = 2
                kwargs['optimal_trajectories'] = None
                kwargs['local_optimization'] = False
            elif cls == DGSM:
                pass
            elif cls == Delta:
                pass
            else:
                raise SAException("Unknown SA class")

            # Generate samples
            param_values = sa.sample(**kwargs)

            # Run model (example)
            results = Ishigami.evaluate(param_values)
            sa.saveResults(results)

            # Perform analysis
            # Si is a Python dict with the keys "S1", "S2", "ST", "S1_conf",
            # "S2_conf", and "ST_conf". The _conf keys store the corresponding
            # confidence intervals, typically with a confidence level of 95%.
            Si = sa.analyze(print_to_console=True)

            # Print the first-order sensitivity indices
            # print Si['S1']



# Copyright (c) 2016  Richard Plevin
# See the https://opensource.org/licenses/MIT for license details.

from pygcam.log import getLogger
from .McsSubcommandABC import McsSubcommandABC

_logger = getLogger(__name__)

def driver(args, tool):
    """
    Run a command for each trialDir or scenarioDir, using str.format to pass
    required args. Possible format args are: projectName, simId, trialNum,
    scenario, simDir, trialDir, and scenarioDir.
    """
    from subprocess import call

    from pygcam.config import getSection

    from ..Database import getDatabase
    from ..error import PygcamMcsUserError
    from ..context import Context
    from .. import util as U

    simId    = args.simId
    command  = args.command
    scenario = args.scenario
    noRun    = args.noRun
    trialStr = args.trials

    projectName = getSection()

    if trialStr:
        trials = U.parseTrialString(trialStr)
    else:
        db = getDatabase()
        count = db.getTrialCount(simId)
        trials = xrange(count)

    # TBD: Add groupName
    context = Context(projectName=projectName, simId=simId, scenario=scenario)
    _logger.info('Running iterator for projectName=%s, simId=%d, scenario=%s, trials=%s, command="%s"',
                 projectName, simId, scenario, trialStr, command)

    # Create a dict to pass to str.format. These are constant across trials.
    argDict = {
        'projectName' : projectName,
        'simId'       : args.simId,
        'scenario'    : args.scenario,
        'expName'     : args.scenario,
    }

    for trialNum in trials:
        argDict['trialNum'] = context.trialNum = trialNum
        argDict['expDir']   = argDict['scenarioDir'] = context.getScenarioDir(create=True)
        argDict['trialDir'] = context.getTrialDir()
        argDict['simDir']   = context.getSimDir()

        try:
            cmd = command.format(**argDict)
        except Exception as e:
            raise PygcamMcsUserError("Bad command format: %s" % e)

        if noRun:
            print(cmd)
        else:
            try:
                call(cmd, shell=True)

            except Exception as e:
                raise PygcamMcsUserError("Failed to run command '%s': %s" % (cmd, e))


class IterateCommand(McsSubcommandABC):
    def __init__(self, subparsers):
        kwargs = {'help' : '''Run a command in each trialDir, or if scenario is given, 
        in each expDir. The following arguments are available for use in the command string,
        specified within curly braces: projectName, simId, trialNum, scenario, expName (a
        legacy alias for scenario), trialDir, scenarioDir, and expDir (a legacy alias for
        scenarioDir). For example, to run the fictional program "foo" in each trialDir for 
        a given set of parameters, you might write:
        gt iterate -s1 -c "foo -s{simId} -t{trialNum} -i{trialDir}/x -o{trialDir}/y/z.txt".'''}
        super(IterateCommand, self).__init__('iterate', subparsers, kwargs)

    def addArgs(self, parser):
        parser.add_argument('-c', '--command', type=str, required=True,
                            help='''A command string to execute for each trial. The following
                            arguments are available for use in the command string, specified
                            within curly braces: projectName, simId, trialNum, scenario, expName, 
                            trialDir, expDir.''')

        parser.add_argument('-n', '--noRun', action='store_true',
                            help="Show the commands that would be executed, but don't run them")

        parser.add_argument('-s', '--simId', type=int, default=1,
                            help='The id of the simulation. Default is 1.')

        parser.add_argument('-S', '--scenario', type=str, default="",
                            help='The name of the scenario')

        parser.add_argument('-t', '--trials', type=str, default=None,
                             help='''Comma separated list of trial or ranges of trials to run. Ex: 1,4,6-10,3.
                             Defaults to running all trials for the given simulation.''')

        return parser   # for auto-doc generation


    def run(self, args, tool):
        driver(args, tool)



# Created on July 5, 2013
#
# @author: Richard Plevin
#
# Copyright (c) 2014. The Regents of the University of California (Regents).
# See the file COPYRIGHT.txt for details.
#
# Note: Useful doc: http://infohost.nmt.edu/tcc/help/pubs/pylxml/web/index.html

from lxml import etree as ET
from pygcam.log import getLogger
from .error import PygcamMcsUserError, PygcamMcsSystemError

_logger = getLogger(__name__)

def pprint(elt):
    ET.tostring(elt, pretty_print=True)

def findAndSave(elt, eltName, cls, myDict, tree=None, testFunc=None, parent=None):
    '''
    Simple helper function to find elements with a given name, create
    a wrapper instance and store them in the given dictionary.
    '''
    for node in elt.iterfind(eltName):
        name = node.get('name')
        obj  = cls(node, tree) if tree else cls(node)

        if parent:
            try:
                obj.setParent(parent)
            except Exception as e:
                raise PygcamMcsSystemError("Call to setParent(%s) on %s failed" % (parent, obj))

        if name in myDict:
            raise PygcamMcsUserError("Attempt to save %s instance with duplicate name '%s'" \
                                     % (eltName, name))

        if testFunc is None or testFunc(obj):
            myDict[name] = obj


class XMLWrapper(object):
    '''
    A simple wrapper around any ElementTree element.
    '''
    def __init__(self, element):
        self.element = element
        self.name    = element.get('name') if element is not None else None

    def getElement(self):
        return self.element

    def getName(self):
        return self.name



#! /usr/bin/env python
'''
.. Created on 4/26/15
.. Copyright (c) 2016 Richard Plevin
   See the https://opensource.org/licenses/MIT for license details.

'''
import os

from .config import getParamAsBoolean
from .constants import LOCAL_XML_NAME
from .log import getLogger
from .query import readQueryResult
from .utils import mkdirs, pathjoin, getBatchDir, getYearCols, printSeries, symlinkOrCopyFile

_logger = getLogger(__name__)

__version__ = "0.1"

PolicyChoices = ['tax', 'subsidy']
DefaultYears = '2020-2050'
DefaultCellulosicCoefficients = "2010:2.057,2015:2.057,2020:2.057,2025:2.039,2030:2.021,2035:2.003,2040:1.986,2045:1.968,2050:1.950,2055:1.932,2060:1.914"

# A list of these is inserted as {constraints} in the metaTemplate
_ConstraintTemplate = '                <constraint year="{year}">{value}</constraint>'

_MetaTemplate = '''<?xml version="1.0" encoding="UTF-8"?>
<scenario>
    <output-meta-data>
        <summary>
        This is a generated constraint file. Edits will be overwritten!
        {summary}
        </summary>
    </output-meta-data>
    <world>
        <region name="{region}">
            <{gcamPolicy} name="{name}">
                <market>{market}</market>
                {preConstraint}
{constraints}
           </{gcamPolicy}>
        </region>
    </world>
</scenario>
'''

DEFAULT_POLICY = 'policy-portfolio-standard'
DEFAULT_REGION = 'USA'
DEFAULT_MARKET = 'USA'

def _generateConstraintXML(name, series, gcamPolicy=DEFAULT_POLICY, policyType=None,
                           region=DEFAULT_REGION, market=DEFAULT_MARKET,
                           preConstraint='', summary=''):

    def genConstraint(year):
        constraint = _ConstraintTemplate.format(year=year, value="{year%s}" % year)
        return constraint

    # Is this the most common case?
    if policyType and not preConstraint:
        preConstraint = '<policyType>{policyType}</policyType>'.format(policyType=policyType)

    constraints = map(genConstraint, series.index)
    constraintText = '\n'.join(constraints)

    template = _MetaTemplate.format(name=name, gcamPolicy=gcamPolicy, policyType=policyType,
                                    region=region, market=market, summary=summary,
                                    preConstraint=preConstraint, constraints=constraintText)

    args = {'year' + col: value for col, value in series.iteritems()}
    xml = template.format(**args)
    return xml


def _saveConstraintFile(xml, dirname, constraintName, policyType, scenario, groupName='',
                        policySrcDir=None): #, fromMCS=False):
    basename = '%s-%s' % (constraintName, policyType)
    constraintFile = basename + '-constraint.xml'
    policyFile     = basename + '.xml'

    fullDirname = pathjoin(dirname, groupName, scenario)
    mkdirs(fullDirname)

    pathname = pathjoin(fullDirname, constraintFile)
    _logger.debug("Generating constraint file: %s", pathname)
    with open(pathname, 'w') as f:
        f.write(xml)

    # TBD: test this
    prefix = '../../../' if groupName else '../../'
    localxml = prefix + LOCAL_XML_NAME

    # ToDo: replace subdir with groupDir?
    #source   = pathjoin(localxml, subdir, scenario, policyFile)

    # TBD: isn't this already handled by addMarketConstraint with baselinePolicy?
    return

    if policySrcDir:
        scenario = policySrcDir
        # parentName = os.path.basename(dirname)
        # grandparentDir = os.path.dirname(os.path.dirname(dirname))
        # fullDirname = pathjoin(grandparentDir, policySrcDir, parentName)
        localxml = pathjoin('../..', policySrcDir, LOCAL_XML_NAME)

    source   = pathjoin(localxml, groupName, scenario, policyFile)
    linkname = pathjoin(fullDirname, policyFile)

    mode = 'Copy' if getParamAsBoolean('GCAM.CopyAllFiles') else 'Link'
    _logger.debug("%sing %s to %s", mode, source, linkname)
    if os.path.lexists(linkname):
        os.remove(linkname)
    symlinkOrCopyFile(source, linkname)

def parseStringPairs(argString, datatype=float):
    """
    Convert a string of comma-separated pairs of colon-delimited values to
    a pandas Series where the first value of each pair is the index name and
    the second value is a float, or the type given.
    """
    import pandas as pd

    pairs = argString.split(',')
    dataDict = {year:datatype(coef) for (year, coef) in map(lambda pair: pair.split(':'), pairs)}
    coefficients = pd.Series(data=dataDict)
    return coefficients

cellEtohConstraintTemplate ='''<?xml version="1.0" encoding="UTF-8"?>
<scenario>
  <output-meta-data>
    <summary>
      Cellulosic ethanol constraints.

      This is a generated constraint file. Edits will be overwritten!
    </summary>
  </output-meta-data>
  <world>
    <region name="USA">
      <policy-portfolio-standard name="cellulosic-etoh-{cellEtohPolicyType}">
        <policyType>{cellEtohPolicyType}</policyType>
        <market>USA</market>
        <min-price year="1975" fillout="1">-1e6</min-price>
        <constraint year="2020">{level2020}</constraint>
        <constraint year="2025">{level2025}</constraint>
        <constraint year="2030">{level2030}</constraint>
        <constraint year="2035">{level2035}</constraint>
        <constraint year="2040">{level2040}</constraint>
        <constraint year="2045">{level2045}</constraint>
        <constraint year="2050">{level2050}</constraint>
      </policy-portfolio-standard>
    </region>
  </world>
</scenario>
'''

US_REGION_QUERY = 'region in ["USA", "United States"]'

# TBD: make region an argument rather than assuming USA

def genBioConstraints(**kwargs):
    import pandas as pd

    #fromMCS = kwargs.get('fromMCS', False)
    resultsDir = kwargs['resultsDir']
    baseline = kwargs['baseline']
    policy = kwargs['policy']
    subdir = kwargs.get('subdir', '')
    defaultLevel = float(kwargs.get('defaultLevel', 0))
    annualLevels = kwargs.get('annualLevels', None)
    biomassPolicyType = kwargs['biomassPolicyType']
    purposeGrownPolicyType = kwargs['purposeGrownPolicyType']
    cellEtohPolicyType = kwargs['cellEtohPolicyType']
    coefficients = parseStringPairs(kwargs.get('coefficients', None) or DefaultCellulosicCoefficients)
    xmlOutputDir = kwargs['xmlOutputDir'] # required

    batchDir = getBatchDir(baseline, resultsDir)

    refinedLiquidsDF = readQueryResult(batchDir, baseline, 'Refined-liquids-production-by-technology', cache=True)
    totalBiomassDF   = readQueryResult(batchDir, baseline, 'Total_biomass_consumption', cache=True)
    purposeGrownDF   = readQueryResult(batchDir, baseline, 'Purpose-grown_biomass_production', cache=True)

    yearCols = getYearCols(kwargs['years'])

    refinedLiquidsUSA = refinedLiquidsDF.query(US_REGION_QUERY)[yearCols]
    totalBiomassUSA   = totalBiomassDF.query(US_REGION_QUERY)[yearCols]
    purposeGrownUSA   = purposeGrownDF.query(US_REGION_QUERY)[yearCols]

    _logger.debug('totalBiomassUSA:\n', totalBiomassUSA)

    cellulosicEtOH  = refinedLiquidsUSA.query('technology == "cellulosic ethanol"')[yearCols]
    if cellulosicEtOH.shape[0] == 0:
        cellulosicEtOH = 0

    _logger.debug('cellulosicEtOH:\n', cellulosicEtOH)
    _logger.debug("Target cellulosic biofuel level %.2f EJ" % defaultLevel)

    desiredCellEtoh = pd.Series(data={year: defaultLevel for year in yearCols})
    if annualLevels:
        annuals = parseStringPairs(annualLevels)
        desiredCellEtoh[annuals.index] = annuals    # override any default values
        _logger.debug("Annual levels set to:", annualLevels)

    _logger.debug("Cell EtOH coefficients:\n", coefficients)

    # Calculate biomass required to meet required level
    deltaCellulose = (desiredCellEtoh - cellulosicEtOH) * coefficients
    _logger.debug('deltaCellulose:\n', deltaCellulose)

    biomassConstraint = totalBiomassUSA.iloc[0] + deltaCellulose.iloc[0]
    _logger.debug('biomassConstraint:\n', biomassConstraint)

    xml = _generateConstraintXML('regional-biomass-constraint', biomassConstraint,
                                 policyType=biomassPolicyType, summary='Regional biomass constraint.')
    _saveConstraintFile(xml, xmlOutputDir, 'regional-biomass', biomassPolicyType, policy,
                        groupName=subdir)#, fromMCS=fromMCS)

    # For switchgrass, we generate a constraint file to adjust purpose-grown biomass
    # by the same amount as the total regional biomass, forcing the change to come from switchgrass.
    if kwargs.get('switchgrass', False):
        constraint = purposeGrownUSA.iloc[0] + deltaCellulose.iloc[0]
    else:
        constraint = purposeGrownUSA.iloc[0]

    xml = _generateConstraintXML('purpose-grown-constraint', constraint, policyType=purposeGrownPolicyType,
                                 summary='Purpose-grown biomass constraint.')
    _saveConstraintFile(xml, xmlOutputDir, 'purpose-grown', purposeGrownPolicyType, policy,
                        groupName=subdir)#, fromMCS=fromMCS)

    # Create dictionary to use for template processing
    xmlArgs = {"level" + year : value for year, value in desiredCellEtoh.iteritems()}
    xmlArgs['cellEtohPolicyType'] = 'subsidy' if cellEtohPolicyType == 'subs' else cellEtohPolicyType

    xml = cellEtohConstraintTemplate.format(**xmlArgs)
    _saveConstraintFile(xml, xmlOutputDir, 'cell-etoh', cellEtohPolicyType, policy,
                        groupName=subdir)#, fromMCS=fromMCS)


def bioMain(args):
    genBioConstraints(**vars(args))


yearConstraintTemplate = '''        <constraint year="{year}">{level}</constraint>'''

fuelConstraintTemplate ='''<?xml version="1.0" encoding="UTF-8"?>
<scenario>
  <output-meta-data>
    <summary>
      Define fuel constraints.
      This is a generated constraint file. Edits will be overwritten!
    </summary>
  </output-meta-data>
  <world>
    <region name="USA">
      <policy-portfolio-standard name="{fuelTag}-{fuelPolicyType}">
        <policyType>{fuelPolicyType}</policyType>
        <market>USA</market>
        <min-price year="1975" fillout="1">-1e6</min-price>
{yearConstraints}
      </policy-portfolio-standard>
    </region>
  </world>
</scenario>
'''

def genDeltaConstraints(**kwargs):
    import pandas as pd

    baseline  = kwargs['baseline']
    policy    = kwargs['policy']
    groupName = kwargs.get('groupName', '')
    fuelTag   = kwargs.get('fuelTag')
    fuelName  = kwargs.get('fuelName')
    resultsDir  = kwargs['resultsDir']
    switchgrass = kwargs.get('switchgrass', False)
    defaultDelta = float(kwargs.get('defaultDelta', 0))
    coefficients = parseStringPairs(kwargs.get('coefficients', None) or DefaultCellulosicCoefficients)
    annualDeltas = kwargs.get('annualDeltas', None)
    xmlOutputDir = kwargs['xmlOutputDir'] # required
    fuelPolicyType = kwargs['fuelPolicyType']
    biomassPolicyType = kwargs.get('biomassPolicyType', None)
    purposeGrownPolicyType = kwargs.get('purposeGrownPolicyType', None)
    policySrcDir = kwargs.get('policySrcDir', False)

    pd.set_option('display.width', None)    # discover width from terminal

    refinedLiquidsQuery = kwargs.get('refinedLiquidsQuery', 'Refined-liquids-production-by-technology')
    totalBiomassQuery   = kwargs.get('totalBiomassQuery',   'Total_biomass_consumption')
    purposeGrownQuery   = kwargs.get('purposeGrownQuery',   'Purpose-grown_biomass_production')

    batchDir = getBatchDir(baseline, resultsDir)
    refinedLiquidsDF = readQueryResult(batchDir, baseline, refinedLiquidsQuery, cache=True)

    yearCols = getYearCols(kwargs['years'])

    combinedQuery = US_REGION_QUERY + ' and technology == "%s"' % fuelName

    fuelBaseline = refinedLiquidsDF.query(combinedQuery)[yearCols]
    if fuelBaseline.shape[0] == 0:
        fuelBaseline = 0
    else:
        printSeries(fuelBaseline, fuelTag, header='fuelBaseline:')

    _logger.debug("Default fuel delta %.2f EJ", defaultDelta)

    deltas = pd.Series(data={year: defaultDelta for year in yearCols})
    if annualDeltas is not None:
        annuals = annualDeltas if isinstance(annualDeltas, pd.Series) else parseStringPairs(annualDeltas)
        deltas.loc[annuals.index] = annuals    # override any default for the given years
        printSeries(deltas, fuelTag, header='annual deltas:')
        #_logger.debug("Annual deltas: %s", deltas)

    # Calculate fuel target after applying deltas
    fuelTargets = fuelBaseline.iloc[0] + deltas
    printSeries(fuelTargets, fuelTag, header='fuelTargets:')

    # Generate annual XML for <constraint year="{year}">{level}</constraint>
    yearConstraints = [yearConstraintTemplate.format(year=year, level=level) for year, level in fuelTargets.iteritems()]

    xmlArgs = {}
    xmlArgs['fuelPolicyType'] = fuelPolicyType
    xmlArgs['fuelTag'] = fuelTag
    xmlArgs['yearConstraints'] = '\n'.join(yearConstraints)

    xml = fuelConstraintTemplate.format(**xmlArgs)

    _saveConstraintFile(xml, xmlOutputDir, fuelTag, fuelPolicyType, policy,
                        groupName=groupName, policySrcDir=policySrcDir)#, fromMCS=fromMCS)

    if switchgrass:
        # Calculate additional biomass required to meet required delta
        deltaCellulose = deltas * coefficients[yearCols]
        printSeries(deltaCellulose, 'cellulose', header='deltaCellulose:')

        totalBiomassDF = readQueryResult(batchDir, baseline, totalBiomassQuery, cache=True)
        totalBiomassUSA = totalBiomassDF.query(US_REGION_QUERY)[yearCols]

        biomassConstraint = totalBiomassUSA.iloc[0] + deltaCellulose.iloc[0]
        printSeries(biomassConstraint, 'regional-biomass', header='biomassConstraint:')

        # For switchgrass, we generate a constraint file to adjust purpose-grown biomass
        # by the same amount as the total regional biomass, forcing the change to come from switchgrass.
        purposeGrownDF = readQueryResult(batchDir, baseline, purposeGrownQuery, cache=True)

        # For some reason, purpose grown results are returned for 1990, 2005, then
        # 2020, 2025, but not 2010 or 2015. So we add any missing columns here.
        missingCols = list(set(yearCols) - set(purposeGrownDF.columns))
        if len(missingCols) > 0:
            purposeGrownDF = pd.concat([purposeGrownDF, pd.DataFrame(columns=missingCols)])

        purposeGrownDF.fillna(0, inplace=True)
        purposeGrownUSA  = purposeGrownDF.query(US_REGION_QUERY)[yearCols]

        xml = _generateConstraintXML('regional-biomass-constraint', biomassConstraint, policyType=biomassPolicyType,
                                     summary='Regional biomass constraint.')

        _saveConstraintFile(xml, xmlOutputDir, 'regional-biomass', biomassPolicyType, policy,
                            groupName=groupName, policySrcDir=policySrcDir)#, fromMCS=fromMCS)

        constraint = purposeGrownUSA.iloc[0] + deltaCellulose.iloc[0]

        xml = _generateConstraintXML('purpose-grown-constraint', constraint, policyType=purposeGrownPolicyType,
                                     summary='Purpose-grown biomass constraint.')

        _saveConstraintFile(xml, xmlOutputDir, 'purpose-grown', purposeGrownPolicyType, policy,
                            groupName=groupName, policySrcDir=policySrcDir)#, fromMCS=fromMCS)



#! /usr/bin/env python
'''
Created on 4/26/15

@author: rjp
'''
from ..subcommand import SubcommandABC

DefaultName  = 'cellulosic ethanol'
DefaultTag   = 'cell-etoh'
DefaultType  = 'subsidy'

class DeltaConstraintsCommand(SubcommandABC):
    def __init__(self, subparsers):
        kwargs = {#'aliases' : ['delta'],
                  'help' : '''Specify incremental values to add to the production of a given fuel,
                              by year, and generate the corresponding constraint file.'''}

        super(DeltaConstraintsCommand, self).__init__('deltaConstraint', subparsers, kwargs)

    def addArgs(self, parser):
        from ..constraints import PolicyChoices, DefaultCellulosicCoefficients, DefaultYears

        parser.add_argument('-c', '--coefficients',
                            help='''A comma-separated string of year:coefficient pairs. This
                            sets the cellulosic ethanol conversion coefficients. Defaults to
                            standard GCAM values: %s.''' % DefaultCellulosicCoefficients)

        parser.add_argument('-b', '--baseline', required=True,
                            help='The baseline on which the policy scenario is based')

        parser.add_argument('-B', '--biomassPolicyType', choices=PolicyChoices, default=None,
                            help='Regional biomass policy type. Default is None.')

        parser.add_argument('-f', '--fuelName', default=DefaultName,
                            help="The fuel to generate constraints for. Default is %s" % DefaultName)

        parser.add_argument('-l', '--defaultDelta', type=float, default=0.0,
                            help='''Default increment to add to each year (EJ). All or individual
                            years values can be set (overriding -l flag values) using the -L flag.''')

        parser.add_argument('-L', '--annualDeltas', default='',
                            help='''Optional production increments by year. Value must be a
                            comma-delimited string of year:level pairs, where level in is EJ.
                            If -l is not used to set default for all years, you must specify
                            values for all years using this option.''')

        # parser.add_argument('-m', '--fromMCS', action='store_true',
        #                      help="Used when calling from gcammcs so correct pathnames are computed.")

        parser.add_argument('-p', '--policy', required=True,
                            help='The policy scenario name')

        parser.add_argument('-P', '--purposeGrownPolicyType', choices=PolicyChoices, default=None,
                             help='Purpose-grown biomass policy type. Default is None.')

        parser.add_argument('-R', '--resultsDir', default='.',
                            help='The parent directory holding the GCAM output workspaces')

        parser.add_argument('-S', '--subdir', default='',
                             help='Sub-directory for local-xml files, if any')

        parser.add_argument('-t', '--fuelTag', default=DefaultTag,
                             help="The fuel tag to generate constraints for. Default is %s" % DefaultTag)

        parser.add_argument('-T', '--policyType', choices=PolicyChoices, default=DefaultType,
                             help='Type of policy to use for the fuel. Default is %s.' % DefaultType)

        parser.add_argument('-x', '--xmlOutputDir',
                             help='''The directory into which to generate XML files.
                             Defaults to policy name in the current directory.''')

        parser.add_argument('-y', '--years', default=DefaultYears,
                            help='''Years to generate constraints for. Must be of the form
                            XXXX-YYYY. Default is "%s"''' % DefaultYears)
        return parser


    def run(self, args, tool):
        from ..constraints import genDeltaConstraints

        genDeltaConstraints(**vars(args))
